@@ -43,28 +43,38 @@
 #include "display.h"
 #include "choose.h"
 #include "errorgui.h"
+#include "gdm-net.h"
 
-static const gchar RCSid[]="$Id: gdm.c,v 1.55 2001/08/02 16:50:29 hp Exp $";
+static const gchar RCSid[]="$Id: gdm.c,v 1.67 2001/08/26 01:56:37 jirka Exp $";
 
 
 /* Local functions */
 static void gdm_config_parse (void);
 static void gdm_local_servers_start (GdmDisplay *d);
+static void gdm_handle_message (GdmConnection *conn,
+				const char *msg,
+				gpointer data);
+static void gdm_handle_user_message (GdmConnection *conn,
+				     const char *msg,
+				     gpointer data);
 static void gdm_daemonify (void);
-static gboolean gdm_slave_socket_handler (GIOChannel *source,
-					  GIOCondition cond,
-					  gpointer data);
 static void gdm_safe_restart (void);
 static void gdm_restart_now (void);
 
 /* Global vars */
-GSList *displays;		/* List of displays managed */
+GSList *displays = NULL;	/* List of displays managed */
+GSList *xservers = NULL;	/* List of x server definitions */
+gint high_display_num = 0;	/* Highest local non-flexi display */
 gint sessions = 0;		/* Number of remote sessions */
+gint flexi_servers = 0;		/* Number of flexi servers */
 sigset_t sysmask;		/* Inherited system signal mask */
 uid_t GdmUserId;		/* Userid under which gdm should run */
 gid_t GdmGroupId;		/* Groupid under which gdm should run */
-int fifofd = -1;		/* fifo fd for the slave communication fifo */
-guint fifo_source = 0;		/* source for the above */
+pid_t extra_process = -1;	/* An extra process.  Used for quickie 
+				   processes, so that they also get whacked */
+
+GdmConnection *fifoconn = NULL; /* Fifo connection */
+GdmConnection *unixconn = NULL; /* UNIX Socket connection */
 
 /* True if the server that was run was in actuallity not specified in the
  * config file.  That is if xdmcp was disabled and no local servers were
@@ -135,6 +145,9 @@ gint  GdmRetryDelay = 0;
 gchar *GdmTimedLogin = NULL;
 gboolean GdmTimedLoginEnable = FALSE;
 gint GdmTimedLoginDelay = 0;
+gchar *GdmStandardXServer = NULL;
+gint  GdmFlexibleXServers = 5;
+gchar *GdmXnest = NULL;
 
 /* set in the main function */
 char **stored_argv = NULL;
@@ -161,6 +174,7 @@ gdm_config_parse (void)
     gchar *bin;
     
     displays = NULL;
+    high_display_num = 0;
 
     if (stat (GDM_CONFIG_FILE, &statbuf) == -1)
 	gdm_error (_("gdm_config_parse: No configuration file: %s. Using defaults."), GDM_CONFIG_FILE);
@@ -229,6 +243,14 @@ gdm_config_parse (void)
     GdmMaxIndirectWait = gnome_config_get_int (GDM_KEY_MAXINDWAIT);    
     GdmPingInterval = gnome_config_get_int (GDM_KEY_PINGINTERVAL);    
 
+    GdmStandardXServer = gnome_config_get_string (GDM_KEY_STANDARD_XSERVER);    
+    if (ve_string_empty (GdmStandardXServer))
+	    GdmStandardXServer = g_strdup ("/usr/bin/X11/X");
+    GdmFlexibleXServers = gnome_config_get_int (GDM_KEY_FLEXIBLE_XSERVERS);    
+    GdmXnest = gnome_config_get_string (GDM_KEY_XNEST);    
+    if (ve_string_empty (GdmXnest))
+	    GdmXnest = NULL;
+
     GdmDebug = gnome_config_get_bool (GDM_KEY_DEBUG);
 
     gnome_config_pop_prefix();
@@ -291,29 +313,103 @@ gdm_config_parse (void)
     if (ve_string_empty (GdmSessDir))
 	gdm_error (_("gdm_config_parse: No sessions directory specified."));
 
+    /* Find server definitions */
+    iter = gnome_config_init_iterator_sections ("=" GDM_CONFIG_FILE "=/");
+    iter = gnome_config_iterator_next (iter, &k, NULL);
+    
+    while (iter) {
+	    if (strncmp (k, "server-", strlen ("server-")) == 0) {
+		    char *section;
+		    GdmXServer *svr = g_new0 (GdmXServer, 1);
+
+		    section = g_strdup_printf ("=" GDM_CONFIG_FILE "=/%s/", k);
+		    gnome_config_push_prefix (section);
+
+		    svr->id = g_strdup (k + strlen ("server-"));
+		    svr->name = gnome_config_get_string (GDM_KEY_SERVER_NAME);
+		    svr->command = gnome_config_get_string
+			    (GDM_KEY_SERVER_COMMAND);
+		    svr->flexible = gnome_config_get_bool
+			    (GDM_KEY_SERVER_FLEXIBLE);
+		    svr->choosable = gnome_config_get_bool
+			    (GDM_KEY_SERVER_CHOOSABLE);
+
+		    if (ve_string_empty (svr->command)) {
+			    gdm_error (_("%s: Empty server command, "
+					 "using standard one."),
+				       "gdm_config_parse");
+			    g_free (svr->command);
+			    svr->command = g_strdup (GdmStandardXServer);
+		    }
+
+		    g_free (section);
+		    gnome_config_pop_prefix ();
+
+		    xservers = g_slist_append (xservers, svr);
+	    }
+
+	    g_free (k);
+
+	    iter = gnome_config_iterator_next (iter, &k, NULL);
+    }
+
+    if (xservers == NULL ||
+	gdm_find_x_server (GDM_STANDARD) == NULL) {
+	    GdmXServer *svr = g_new0 (GdmXServer, 1);
+
+	    svr->id = g_strdup (GDM_STANDARD);
+	    svr->name = g_strdup ("Standard server");
+	    svr->command = g_strdup (GdmStandardXServer);
+	    svr->flexible = TRUE;
+	    svr->choosable = TRUE;
+
+	    xservers = g_slist_append (xservers, svr);
+    }
 
     /* Find local X server definitions */
     iter = gnome_config_init_iterator ("=" GDM_CONFIG_FILE "=/" GDM_KEY_SERVERS);
     iter = gnome_config_iterator_next (iter, &k, &v);
     
     while (iter) {
+	    if (isdigit (*k)) {
+		    int disp_num = atoi (k);
+		    GdmDisplay *disp = gdm_server_alloc (disp_num, v);
+		    if (disp == NULL) {
+			    g_free (k);
+			    g_free (v);
+			    iter = gnome_config_iterator_next (iter, &k, &v);
+			    continue;
+		    }
+		    displays = g_slist_append (displays, disp);
+		    if (disp_num > high_display_num)
+			    high_display_num = disp_num;
+	    } else {
+		    gdm_info (_("gdm_config_parse: Invalid server line in config file. Ignoring!"));
+	    }
+	    g_free (k);
+	    g_free (v);
 
-	if (isdigit (*k))
-	    displays = g_slist_append (displays, gdm_server_alloc (atoi (k), v));
-	else
-	    gdm_info (_("gdm_config_parse: Invalid server line in config file. Ignoring!"));
-
-	iter = gnome_config_iterator_next (iter, &k, &v);
+	    iter = gnome_config_iterator_next (iter, &k, &v);
     }
 
-    if (! displays && ! GdmXdmcp) {
+    if (displays == NULL && ! GdmXdmcp) {
+	    char *server = NULL;
+	    if (access (GdmStandardXServer, X_OK) == 0) {
+		    server = GdmStandardXServer;
+	    } else if (access ("/usr/bin/X11/X", X_OK) == 0) {
+		    server = "/usr/bin/X11/X";
+	    }
 	    /* yay, we can add a backup emergency server */
-	    if (access ("/usr/bin/X11/X", X_OK) == 0) {
-		    gdm_error (_("gdm_config_parse: Xdmcp disabled and no local servers defined. Adding /usr/bin/X11/X on :0 to allow configuration!"));
+	    if (server != NULL) {
+		    int num = gdm_get_free_display (0 /* start */,
+						    0 /* server uid */);
+		    gdm_error (_("%s: Xdmcp disabled and no local servers defined. Adding /usr/bin/X11/X on :%d to allow configuration!"),
+			       "gdm_config_parse",
+			       num);
 
 		    gdm_emergency_server = TRUE;
 		    displays = g_slist_append
-			    (displays, gdm_server_alloc (0, "/usr/bin/X11/X"));
+			    (displays, gdm_server_alloc (num, server));
 		    /* ALWAYS run the greeter and don't log anyone in,
 		     * this is just an emergency session */
 		    g_free (GdmAutomaticLogin);
@@ -410,6 +506,25 @@ gdm_config_parse (void)
     gdm_verify_check ();
 }
 
+/* If id == NULL, then get the first X server */
+GdmXServer *
+gdm_find_x_server (const char *id)
+{
+	GSList *li;
+
+	if (xservers == NULL)
+		return NULL;
+
+	if (id == NULL)
+		return xservers->data;
+
+	for (li = xservers; li != NULL; li = li->next) {
+		GdmXServer *svr = li->data;
+		if (strcmp (svr->id, id) == 0)
+			return svr;
+	}
+	return NULL;
+}
 
 /**
  * gdm_daemonify:
@@ -474,7 +589,6 @@ gdm_local_servers_start (GdmDisplay *d)
 		/* only the first local display gets autologged in */
 		gdm_first_login = FALSE;
 	}
-
     }
 }
 
@@ -483,6 +597,7 @@ final_cleanup (void)
 {
 	GSList *list;
 	sigset_t mask;
+	gchar *path;
 
 	gdm_debug ("final_cleanup");
 
@@ -490,12 +605,33 @@ final_cleanup (void)
 	sigaddset (&mask, SIGCHLD);
 	sigprocmask (SIG_BLOCK, &mask, NULL); 
 
+	if (extra_process > 1) {
+		/* we sigterm extra processes, and we
+		 * don't wait */
+		kill (extra_process, SIGTERM);
+		extra_process = -1;
+	}
+
 	list = g_slist_copy (displays);
 	g_slist_foreach (list, (GFunc) gdm_display_unmanage, NULL);
 	g_slist_free (list);
 
+	/* Close stuff */
+
 	gdm_xdmcp_close ();
-	gdm_fifo_close ();
+	gdm_connection_close (fifoconn);
+	fifoconn = NULL;
+	gdm_connection_close (unixconn);
+	unixconn = NULL;
+
+	/* Unlink the connections */
+
+	path = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
+	unlink (path);
+	g_free (path);
+
+	unlink (GDM_SUP_SOCKET);
+
 
 	closelog();
 	unlink (GdmPidFile);
@@ -505,7 +641,6 @@ static gboolean
 deal_with_x_crashes (GdmDisplay *d)
 {
     gboolean just_abort = FALSE;
-    char *msg;
 
     if ( ! d->failsafe_xserver &&
 	 ! ve_string_empty (GdmFailsafeXServer)) {
@@ -543,19 +678,35 @@ deal_with_x_crashes (GdmDisplay *d)
 	    if (configurators[i] != NULL) {
 		    strcpy (tempname, "/tmp/gdm-X-failed-XXXXXX");
 		    tempfd = mkstemp (tempname);
-		    close (tempfd);
+		    if (tempfd >= 0) {
+			    close (tempfd);
 
-		    gdm_info (_("deal_with_x_crashes: Running the "
-				"XKeepsCrashing script"));
+			    gdm_info (_("deal_with_x_crashes: Running the "
+					"XKeepsCrashing script"));
 
-		    pid = fork ();
+			    extra_process = pid = fork ();
+		    } else {
+			    /* no forking, we're screwing this */
+			    pid = -1;
+		    }
 	    } else {
 		    tempfd = -1;
 		    /* no forking, we're screwing this */
 		    pid = -1;
 	    }
 	    if (pid == 0) {
-		    char *argv[9];
+		    char *argv[10];
+		    char *xlog = g_strconcat (GdmLogDir, "/", d->name, ".log", NULL);
+		    int ii;
+
+		    for (ii = 0; ii < sysconf (_SC_OPEN_MAX); ii++)
+			    close (ii);
+
+		    /* No error checking here - if it's messed the best response
+		    * is to ignore & try to continue */
+		    open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
+		    open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
+		    open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
 
 		    argv[0] = GdmXKeepsCrashing;
 		    argv[1] = configurators[i];
@@ -565,9 +716,7 @@ deal_with_x_crashes (GdmDisplay *d)
 				"up correctly.  You will need to log in on a "
 				"console and rerun the X configuration "
 				"program.  Then restart GDM.");
-		    argv[4] = _("I cannot start the X server (your graphical "
-				"interface).  It is likely that it is not set "
-				"up correctly.  Would you like me to try to "
+		    argv[4] = _("Would you like me to try to "
 				"run the X configuration program?  Note that "
 				"you will need the root password for this.");
 		    argv[5] = _("Please type in the root (privilaged user) "
@@ -576,7 +725,20 @@ deal_with_x_crashes (GdmDisplay *d)
 				"again.");
 		    argv[7] = _("I will disable this X server for now.  "
 				"Restart GDM when it is configured correctly.");
-		    argv[8] = NULL;
+		    argv[8] = _("I cannot start the X server (your graphical "
+				"interface).  It is likely that it is not set "
+				"up correctly.  Would you like to view the "
+				"X server output to diagnose the problem?");
+		    argv[9] = NULL;
+
+		    /* unset DISPLAY and XAUTHORITY if they exist
+		     * so that gdialog (if used) doesn't get confused */
+		    ve_unsetenv ("DISPLAY");
+		    ve_unsetenv ("XAUTHORITY");
+
+		    ve_setenv ("XLOG", xlog, TRUE);
+		    ve_setenv ("BINDIR", EXPANDED_BINDIR, TRUE);
+		    ve_setenv ("SBINDIR", EXPANDED_SBINDIR, TRUE);
 
 		    execv (argv[0], argv);
 	
@@ -585,13 +747,16 @@ deal_with_x_crashes (GdmDisplay *d)
 	    } else if (pid > 0) {
 		    int status;
 		    waitpid (pid, &status, 0);
+
+		    extra_process = -1;
+
 		    if (WIFEXITED (status) &&
 			WEXITSTATUS (status) == 0) {
 			    /* Yay, the user wants to try again, so
 			     * here we go */
 			    return TRUE;
 		    } else if (WIFEXITED (status) &&
-			     WEXITSTATUS (status) == 32) {
+			       WEXITSTATUS (status) == 32) {
 			    /* We couldn't run the script, just drop through */
 			    ;
 		    } else {
@@ -611,9 +776,7 @@ deal_with_x_crashes (GdmDisplay *d)
     /* if we have "open" we can talk to the user, not as user
      * friendly as the above script, but getting there */
     if ( ! just_abort &&
-	access ("/usr/bin/open", X_OK) == 0) {
-	    char *dialog; /* do we have dialog?*/
-
+	access (EXPANDED_SBINDIR "/gdmopen", X_OK) == 0) {
 	    /* Shit if we knew what the program was to tell the user,
 	     * the above script would have been defined and we'd run
 	     * it for them */
@@ -622,38 +785,14 @@ deal_with_x_crashes (GdmDisplay *d)
 			    "up correctly.  You will need to log in on a "
 			    "console and rerun the X configuration "
 			    "program.  Then restart GDM.");
-
-	    dialog = gnome_is_program_in_path ("dialog");
-	    if (dialog == NULL)
-		    dialog = gnome_is_program_in_path ("gdialog");
-	    if (dialog != NULL) {
-		    char *command = 
-			    g_strdup_printf
-			    ("/usr/bin/open -s -w -- /bin/sh -c 'clear ; "
-			     "%s --msgbox \"%s\" 10 70 ; clear'",
-			     dialog,
-			     error);
-		    system (command);
-		    g_free (command);
-		    g_free (dialog);
-	    } else {
-		    char *command = 
-			    g_strdup_printf
-			    ("/usr/bin/open -s -w -- /bin/sh -c 'clear ; "
-			     "echo \"%s\" 10 70 ; read ; clear'",
-			     error);
-		    system (command);
-		    g_free (command);
-	    }
+	    gdm_text_message_dialog (error);
     } /* else {
        * At this point .... screw the user, we don't know how to
        * talk to him.  He's on some 'l33t system anyway, so syslog
        * reading will do him good 
        * } */
 
-    msg = g_strdup_printf (_("Failed to start X server several times in a short time period; disabling display %s"), d->name);
-    gdm_error (msg);
-    g_free (msg);
+    gdm_error (_("Failed to start X server several times in a short time period; disabling display %s"), d->name);
 
     return FALSE;
 }
@@ -756,7 +895,7 @@ gdm_cleanup_children (void)
 	    status = DISPLAY_REMANAGE;
     }
 
-    if (d->type != TYPE_LOCAL &&
+    if ( ! d->console &&
 	(status == DISPLAY_RESTARTGDM ||
 	 status == DISPLAY_REBOOT ||
 	 status == DISPLAY_HALT)) {
@@ -840,10 +979,18 @@ gdm_cleanup_children (void)
 	break;
 
     case DISPLAY_XFAILED:       /* X sucks */
+	/* inform about error if needed */
+	if (d->socket_conn != NULL) {
+		GdmConnection *conn = d->socket_conn;
+		d->socket_conn = NULL;
+		gdm_connection_set_close_notify (conn, NULL, NULL);
+		gdm_connection_write (conn, "ERROR 3 X failed\n");
+	}
+
 	if (gdm_restart_mode)
 		gdm_safe_restart ();
 
-	/* in remote case just drop to _REMANAGE */
+	/* in remote/flexi case just drop to _REMANAGE */
 	if (d->type == TYPE_LOCAL) {
 		time_t now = time (NULL);
 		d->x_faileds ++;
@@ -877,6 +1024,14 @@ gdm_cleanup_children (void)
     default:
 	gdm_debug ("gdm_child_action: Slave process returned %d", status);
 
+	/* inform about error if needed */
+	if (d->socket_conn != NULL) {
+		GdmConnection *conn = d->socket_conn;
+		d->socket_conn = NULL;
+		gdm_connection_set_close_notify (conn, NULL, NULL);
+		gdm_connection_write (conn, "ERROR 2 Startup errors\n");
+	}
+
 	if (gdm_restart_mode)
 		gdm_safe_restart ();
 	
@@ -885,7 +1040,7 @@ gdm_cleanup_children (void)
 		if ( ! gdm_display_manage (d))
 			gdm_display_unmanage (d);
 	/* Remote displays will send a request to be managed */
-	} else if (d->type == TYPE_XDMCP) {
+	} else /* TYPE_XDMCP, TYPE_FLEXI, TYPE_FLEXI_XNEST */ {
 		gdm_display_unmanage (d);
 	}
 	
@@ -985,39 +1140,27 @@ store_argv (int argc, char *argv[])
 }
 
 static void
-create_fifo (void)
+create_connections (void)
 {
-	gchar *fifopath;
-	GIOChannel *fifochan;
+	gchar *path;
 
-	fifopath = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
+	path = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
+	fifoconn = gdm_connection_open_fifo (path, 0660);
+	g_free (path);
 
-	unlink (fifopath);
+	if (fifoconn != NULL)
+		gdm_connection_set_handler (fifoconn,
+					    gdm_handle_message,
+					    NULL /* data */,
+					    NULL /* destroy_notify */);
 
-	if (mkfifo (fifopath, 0660) < 0) {
-		gdm_fail (_("%s: Could not make FIFO"),
-			  "create_fifo");
-		return;
-	}
+	unixconn = gdm_connection_open_unix (GDM_SUP_SOCKET, 0666);
 
-	fifofd = open (fifopath, O_RDWR); /* Open with write to avoid EOF */
-
-	if (fifofd < 0) {
-		gdm_fail (_("%s: Could not open FIFO"),
-			  "create_fifo");
-		return;
-	}
-
-	chmod (fifopath, 0660);
-
-	g_free (fifopath);
-
-	fifochan = g_io_channel_unix_new (fifofd);
-	fifo_source = g_io_add_watch_full
-		(fifochan, G_PRIORITY_DEFAULT,
-		 G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
-		 gdm_slave_socket_handler, NULL, NULL);
-	g_io_channel_unref (fifochan);
+	if (unixconn != NULL)
+		gdm_connection_set_handler (unixconn,
+					    gdm_handle_user_message,
+					    NULL /* data */,
+					    NULL /* destroy_notify */);
 }
 
 int 
@@ -1048,8 +1191,11 @@ main (int argc, char *argv[])
     }
 
     /* XDM compliant error message */
-    if (getuid() != 0)
-	gdm_fail (_("Only root wants to run gdm\n"));
+    if (getuid() != 0) {
+	    /* make sure the pid file doesn't get wiped */
+	    GdmPidFile = NULL;
+	    gdm_fail (_("Only root wants to run gdm\n"));
+    }
 
 
     /* Initialize runtime environment */
@@ -1073,8 +1219,11 @@ main (int argc, char *argv[])
 
         if (pf != NULL &&
 	    fscanf (pf, "%d", &pidv) == 1 &&
-	    kill (pidv, 0) != -1)
+	    kill (pidv, 0) != -1) {
+		/* make sure the pid file doesn't get wiped */
+		GdmPidFile = NULL;
 		gdm_fail (_("gdm already running. Aborting!"));
+	}
 
 	if (pf != NULL)
 		fclose (pf);
@@ -1141,7 +1290,7 @@ main (int argc, char *argv[])
     if (GdmXdmcp)
 	gdm_xdmcp_init();
 
-    create_fifo ();
+    create_connections ();
 
     /* Start local X servers */
     g_slist_foreach (displays, (GFunc) gdm_local_servers_start, NULL);
@@ -1281,15 +1430,27 @@ gdm_quit (void)
 }
 
 static void
-handle_message (const char *msg)
+gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 {
-	gdm_debug ("Handeling message: '%s'", msg);
+	/* Evil!, all this for debugging? */
+	if (GdmDebug) {
+		if (strncmp (msg, GDM_SOP_COOKIE " ",
+			     strlen (GDM_SOP_COOKIE " ")) == 0) {
+			char *s = g_strndup
+				(msg, strlen (GDM_SOP_COOKIE " XXXX XX"));
+			/* cut off most of the cookie for "security" */
+			gdm_debug ("Handeling message: '%s...'", s);
+			g_free (s);
+		} else {
+			gdm_debug ("Handeling message: '%s'", msg);
+		}
+	}
 
 	if (strncmp (msg, GDM_SOP_CHOSEN " ",
 		     strlen (GDM_SOP_CHOSEN " ")) == 0) {
 		gdm_choose_data (msg);
 	} else if (strncmp (msg, GDM_SOP_XPID " ",
-		     strlen (GDM_SOP_XPID " ")) == 0) {
+		            strlen (GDM_SOP_XPID " ")) == 0) {
 		GdmDisplay *d;
 		long slave_pid, pid;
 
@@ -1307,7 +1468,7 @@ handle_message (const char *msg)
 			kill (slave_pid, SIGUSR2);
 		}
 	} else if (strncmp (msg, GDM_SOP_SESSPID " ",
-		     strlen (GDM_SOP_SESSPID " ")) == 0) {
+		            strlen (GDM_SOP_SESSPID " ")) == 0) {
 		GdmDisplay *d;
 		long slave_pid, pid;
 
@@ -1325,7 +1486,7 @@ handle_message (const char *msg)
 			kill (slave_pid, SIGUSR2);
 		}
 	} else if (strncmp (msg, GDM_SOP_GREETPID " ",
-		     strlen (GDM_SOP_GREETPID " ")) == 0) {
+		            strlen (GDM_SOP_GREETPID " ")) == 0) {
 		GdmDisplay *d;
 		long slave_pid, pid;
 
@@ -1343,7 +1504,7 @@ handle_message (const char *msg)
 			kill (slave_pid, SIGUSR2);
 		}
 	} else if (strncmp (msg, GDM_SOP_CHOOSERPID " ",
-		     strlen (GDM_SOP_CHOOSERPID " ")) == 0) {
+		            strlen (GDM_SOP_CHOOSERPID " ")) == 0) {
 		GdmDisplay *d;
 		long slave_pid, pid;
 
@@ -1361,7 +1522,7 @@ handle_message (const char *msg)
 			kill (slave_pid, SIGUSR2);
 		}
 	} else if (strncmp (msg, GDM_SOP_LOGGED_IN " ",
-		     strlen (GDM_SOP_LOGGED_IN " ")) == 0) {
+		            strlen (GDM_SOP_LOGGED_IN " ")) == 0) {
 		GdmDisplay *d;
 		long slave_pid;
 		int logged_in;
@@ -1386,61 +1547,458 @@ handle_message (const char *msg)
 			/* send ack */
 			kill (slave_pid, SIGUSR2);
 		}
+	} else if (strncmp (msg, GDM_SOP_DISP_NUM " ",
+		            strlen (GDM_SOP_DISP_NUM " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid;
+		int disp_num;
+
+		if (sscanf (msg, GDM_SOP_DISP_NUM " %ld %d",
+			    &slave_pid, &disp_num) != 2)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			g_free (d->name);
+			d->name = g_strdup_printf (":%d", disp_num);
+			d->dispnum = disp_num;
+			gdm_debug ("Got DISP_NUM == %d", disp_num);
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
+	} else if (strncmp (msg, GDM_SOP_VT_NUM " ",
+		            strlen (GDM_SOP_VT_NUM " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid;
+		int vt_num;
+
+		if (sscanf (msg, GDM_SOP_VT_NUM " %ld %d",
+			    &slave_pid, &vt_num) != 2)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+#ifdef __linux__
+			d->vt = vt_num;
+#endif
+			gdm_debug ("Got VT_NUM == %d", vt_num);
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
+	} else if (strncmp (msg, GDM_SOP_LOGIN " ",
+		            strlen (GDM_SOP_LOGIN " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid;
+		char *p;
+
+		if (sscanf (msg, GDM_SOP_LOGIN " %ld",
+			    &slave_pid) != 1)
+			return;
+		p = strchr (msg, ' ');
+		if (p != NULL)
+			p = strchr (p+1, ' ');
+		if (p == NULL)
+			return;
+
+		p++;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			g_free (d->login);
+			d->login = g_strdup (p);
+			gdm_debug ("Got LOGIN == %s", p);
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
+	} else if (strncmp (msg, GDM_SOP_COOKIE " ",
+		            strlen (GDM_SOP_COOKIE " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid;
+		char *p;
+
+		if (sscanf (msg, GDM_SOP_COOKIE " %ld",
+			    &slave_pid) != 1)
+			return;
+		p = strchr (msg, ' ');
+		if (p != NULL)
+			p = strchr (p+1, ' ');
+		if (p == NULL)
+			return;
+
+		p++;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			g_free (d->cookie);
+			d->cookie = g_strdup (p);
+			gdm_debug ("Got COOKIE == <secret>");
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
+	} else if (strncmp (msg, GDM_SOP_FLEXI_ERR " ",
+		            strlen (GDM_SOP_FLEXI_ERR " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid;
+		int err;
+
+		if (sscanf (msg, GDM_SOP_FLEXI_ERR " %ld %d",
+			    &slave_pid, &err) != 2)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			char *error = NULL;
+			GdmConnection *conn = d->socket_conn;
+			d->socket_conn = NULL;
+
+			if (conn != NULL)
+				gdm_connection_set_close_notify (conn,
+								 NULL, NULL);
+
+			if (err == 3)
+				error = "ERROR 3 X failed\n";
+			else if (err == 4)
+				error = "ERROR 4 X too busy\n";
+			else if (err == 5)
+				error = "ERROR 5 Xnest can't connect\n";
+			else
+				error = "ERROR 999 Unknown error\n";
+			if (conn != NULL)
+				gdm_connection_write (conn, error);
+			
+			gdm_debug ("Got FLEXI_ERR == %d", err);
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
+	} else if (strncmp (msg, GDM_SOP_FLEXI_OK " ",
+		            strlen (GDM_SOP_FLEXI_OK " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid;
+
+		if (sscanf (msg, GDM_SOP_FLEXI_OK " %ld",
+			    &slave_pid) != 1)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			GdmConnection *conn = d->socket_conn;
+			d->socket_conn = NULL;
+
+			if (conn != NULL) {
+				char *msg = g_strdup_printf
+					("OK %s\n", d->name);
+				gdm_connection_set_close_notify (conn,
+								 NULL, NULL);
+				if ( ! gdm_connection_write (conn, msg))
+					gdm_display_unmanage (d);
+			}
+			
+			gdm_debug ("Got FLEXI_OK");
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
 	} else if (strcmp (msg, GDM_SOP_SOFT_RESTART) == 0) {
 		gdm_restart_mode = TRUE;
 		gdm_safe_restart ();
 	}
 }
 
-static gboolean
-gdm_slave_socket_handler (GIOChannel *source,
-			  GIOCondition cond,
-			  gpointer data)
+/* extract second word and the rest of the string */
+static void
+extract_dispname_xauthfile (const char *msg, char **dispname, char **xauthfile)
 {
-	gchar buf[PIPE_SIZE];
-	char *p;
-	gint len;
+	const char *p;
+	char *pp;
 
-	if (cond != G_IO_IN) 
-		return TRUE;
+	*dispname = NULL;
+	*xauthfile = NULL;
 
-	if (g_io_channel_read (source, buf, sizeof (buf) - 1, &len)
-	    != G_IO_ERROR_NONE)
-		return TRUE;
+	p = strchr (msg, ' ');
+	if (p == NULL)
+		return;
 
-	if (len <= 0)
-		return TRUE;
+	while (*p == ' ')
+		p++;
 
-	/* null terminate as the string is NOT */
-	buf[len] = '\0';
+	*dispname = g_strdup (p);
 
-	p = strtok (buf, "\n");
-	while (p != NULL) {
-		handle_message (p);
-		p = strtok (NULL, "\n");
+	p = strchr (p, ' ');
+	if (p == NULL) {
+		return;
 	}
 
-	return TRUE;
+	while (*p == ' ')
+		p++;
+
+	*xauthfile = g_strstrip (g_strdup (p));
+
+	pp = strchr (*dispname, ' ');
+	if (pp != NULL)
+		*pp = '\0';
 }
 
-void
-gdm_fifo_close (void)
+static void
+close_conn (gpointer data)
 {
-	gchar *fifopath;
+	GdmDisplay *disp = data;
 
-	if (fifo_source > 0) {
-		g_source_remove (fifo_source);
-		fifo_source = 0;
+	/* We still weren't finished, so we want to whack this display */
+	if (disp->socket_conn != NULL) {
+		disp->socket_conn = NULL;
+		gdm_display_unmanage (disp);
+	}
+}
+
+static GdmDisplay *
+find_display (const char *name)
+{
+	GSList *li;
+
+	for (li = displays; li != NULL; li = li->next) {
+		GdmDisplay *disp = li->data;
+		if (disp->name != NULL &&
+		    strcmp (disp->name, name) == 0)
+			return disp;
+	}
+	return NULL;
+}
+
+static void
+handle_flexi_server (GdmConnection *conn, int type, const char *server,
+		     const char *xnest_disp, const char *xnest_auth_file)
+{
+	GdmDisplay *display;
+	char *bin;
+
+	gdm_debug ("server: '%s'", server);
+
+	if (flexi_servers >= GdmFlexibleXServers) {
+		gdm_connection_write (conn,
+				      "ERROR 1 No more flexi servers\n");
+		return;
 	}
 
-	if (fifofd > 0) {
-		close (fifofd);
-		fifofd = -1;
+	bin = ve_first_word (ve_sure_string (server));
+	if (server == NULL ||
+	    access (bin, X_OK) != 0) {
+		g_free (bin);
+		gdm_connection_write (conn,
+				      "ERROR 6 No server binary\n");
+		return;
+	}
+	g_free (bin);
+
+	display = gdm_server_alloc (-1, server);
+	if (display == NULL) {
+		gdm_connection_write (conn,
+				      "ERROR 2 Startup errors\n");
+		return;
 	}
 
-	fifopath = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
-	unlink (fifopath);
-	g_free (fifopath);
+	if (type == TYPE_FLEXI_XNEST) {
+		GdmDisplay *parent;
+		char *disp, *p;
+		g_assert (xnest_disp != NULL);
+
+		disp = g_strdup (xnest_disp);
+		/* whack the screen info */
+		p = strchr (disp, ':');
+		if (p != NULL)
+			p = strchr (p+1, '.');
+		if (p != NULL)
+			*p = '\0';
+		/* if it's on one of the console displays we started,
+		 * it's on the console, else it's not (it could be but
+		 * we aren't sure and we don't want to be fooled) */
+		parent = find_display (disp);
+		if (/* paranoia */xnest_disp[0] == ':' &&
+		    parent != NULL &&
+		    parent->console)
+			display->console = TRUE;
+		else
+			display->console = FALSE;
+		g_free (disp);
+	}
+
+	flexi_servers ++;
+
+	display->type = type;
+	display->socket_conn = conn;
+	display->xnest_disp = g_strdup (xnest_disp);
+	display->xnest_auth_file = g_strdup (xnest_auth_file);
+	gdm_connection_set_close_notify (conn, display, close_conn);
+	displays = g_slist_append (displays, display);
+	if ( ! gdm_display_manage (display)) {
+		gdm_display_unmanage (display);
+		gdm_connection_write (conn,
+				      "ERROR 2 Startup errors\n");
+		return;
+	}
+	/* Now we wait for the server to start up (or not) */
+}
+
+static void
+gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
+{
+	gdm_debug ("Handeling user message: '%s'", msg);
+
+	if (strncmp (msg, GDM_SUP_AUTH_LOCAL " ",
+		     strlen (GDM_SUP_AUTH_LOCAL " ")) == 0) {
+		GSList *li;
+		char *cookie = g_strdup
+			(&msg[strlen (GDM_SUP_AUTH_LOCAL " ")]);
+		g_strstrip (cookie);
+		if (strlen (cookie) != 16*2) /* 16 bytes in hex form */ {
+			/* evil, just whack the connection in this case */
+			gdm_connection_write (conn,
+					      "ERROR 100 Not authenticated\n");
+			gdm_connection_close (conn);
+			g_free (cookie);
+			return;
+		}
+		/* check if cookie matches one of the console displays */
+		for (li = displays; li != NULL; li = li->next) {
+			GdmDisplay *disp = li->data;
+			if (disp->console &&
+			    disp->cookie != NULL &&
+			    strcmp (disp->cookie, cookie) == 0) {
+				g_free (cookie);
+				GDM_CONNECTION_SET_USER_FLAG
+					(conn, GDM_SUP_FLAG_AUTHENTICATED);
+				gdm_connection_write (conn, "OK\n");
+				return;
+			}
+		}
+
+		/* Hmmm, perhaps this is better defined behaviour */
+		GDM_CONNECTION_UNSET_USER_FLAG
+			(conn, GDM_SUP_FLAG_AUTHENTICATED);
+		gdm_connection_write (conn, "ERROR 100 Not authenticated\n");
+		g_free (cookie);
+	} else if (strcmp (msg, GDM_SUP_FLEXI_XSERVER) == 0) {
+		/* Only allow locally authenticated connections */
+		if ( ! (gdm_connection_get_user_flags (conn) &
+			GDM_SUP_FLAG_AUTHENTICATED)) {
+			gdm_info (_("Flexible server request denied: "
+				    "Not authenticated"));
+			gdm_connection_write (conn,
+					      "ERROR 100 Not authenticated\n");
+			return;
+		}
+		handle_flexi_server (conn, TYPE_FLEXI, GdmStandardXServer,
+				     NULL, NULL);
+	} else if (strncmp (msg, GDM_SUP_FLEXI_XSERVER " ",
+		            strlen (GDM_SUP_FLEXI_XSERVER " ")) == 0) {
+		char *name;
+		const char *command = NULL;
+		GdmXServer *svr;
+
+		/* Only allow locally authenticated connections */
+		if ( ! (gdm_connection_get_user_flags (conn) &
+			GDM_SUP_FLAG_AUTHENTICATED)) {
+			gdm_info (_("Flexible server request denied: "
+				    "Not authenticated"));
+			gdm_connection_write (conn,
+					      "ERROR 100 Not authenticated\n");
+			return;
+		}
+
+		name = g_strdup (&msg[strlen (GDM_SUP_FLEXI_XSERVER " ")]);
+		g_strstrip (name);
+		if (ve_string_empty (name)) {
+			g_free (name);
+			name = g_strdup (GDM_STANDARD);
+		}
+
+		svr = gdm_find_x_server (name);
+		g_free (name);
+		if (svr == NULL) {
+			/* Don't print the name to syslog as it might be
+			 * long and dangerous */
+			gdm_error (_("Unknown server type requested, using "
+				     "standard server."));
+			command = GdmStandardXServer;
+		} else if ( ! svr->flexible) {
+			gdm_error (_("Requested server %s not allowed to be "
+				     "used for flexible servers, using "
+				     "standard server."), name);
+			command = GdmStandardXServer;
+		} else {
+			command = svr->command;
+		}
+
+		handle_flexi_server (conn, TYPE_FLEXI, command, NULL, NULL);
+	} else if (strncmp (msg, GDM_SUP_FLEXI_XNEST " ",
+		            strlen (GDM_SUP_FLEXI_XNEST " ")) == 0) {
+		char *dispname = NULL, *xauthfile = NULL;
+
+		extract_dispname_xauthfile (msg, &dispname, &xauthfile);
+		if (dispname == NULL) {
+			/* Something bogus is going on, so just whack the
+			 * connection */
+			g_free (xauthfile);
+			gdm_connection_close (conn);
+			return;
+		}
+		
+		handle_flexi_server (conn, TYPE_FLEXI_XNEST, GdmXnest,
+				     dispname, xauthfile);
+
+		g_free (dispname);
+		g_free (xauthfile);
+	} else if (strcmp (msg, GDM_SUP_CONSOLE_SERVERS) == 0) {
+		GSList *li;
+		const char *sep = " ";
+		gdm_connection_write (conn, "OK");
+		for (li = displays; li != NULL; li = li->next) {
+			GdmDisplay *disp = li->data;
+			if ( ! disp->console)
+				continue;
+			gdm_connection_write (conn, sep);
+			sep = ";";
+			gdm_connection_write (conn,
+					      ve_sure_string (disp->name));
+			gdm_connection_write (conn, ",");
+			gdm_connection_write (conn,
+					      ve_sure_string (disp->login));
+			gdm_connection_write (conn, ",");
+			if (disp->type == TYPE_FLEXI_XNEST) {
+				gdm_connection_write
+					(conn,
+					 ve_sure_string (disp->xnest_disp));
+			} else {
+				char *vt = g_strdup_printf ("%d",
+#ifdef __linux__
+							    disp->vt
+#else
+							    -1
+#endif
+							    );
+				gdm_connection_write (conn, vt);
+				g_free (vt);
+			}
+		}
+		gdm_connection_write (conn, "\n");
+	} else if (strcmp (msg, GDM_SUP_VERSION) == 0) {
+		gdm_connection_write (conn, "GDM " VERSION "\n");
+	} else if (strcmp (msg, GDM_SUP_CLOSE) == 0) {
+		gdm_connection_close (conn);
+	} else {
+		gdm_connection_write (conn, "ERROR 0 Not implemented\n");
+		gdm_connection_close (conn);
+	}
 }
 
 /* EOF */
