@@ -24,13 +24,20 @@
 #include <X11/Xmd.h>
 #include <X11/Xauth.h>
 
-
 #define STX 0x2			/* Start of txt */
 #define BEL 0x7			/* Bell, used to interrupt login for
 				 * say timed login or something similar */
 
 #define TYPE_LOCAL 1		/* Local X server */
 #define TYPE_XDMCP 2		/* Remote display */
+#define TYPE_FLEXI 3		/* Local Flexi X server */
+#define TYPE_FLEXI_XNEST 4	/* Local Flexi Xnest server */
+
+#define SERVER_IS_LOCAL(d) (d->type == TYPE_LOCAL || \
+			    d->type == TYPE_FLEXI || \
+			    d->type == TYPE_FLEXI_XNEST)
+#define SERVER_IS_FLEXI(d) (d->type == TYPE_FLEXI || \
+			    d->type == TYPE_FLEXI_XNEST)
 
 #define SERVER_SUCCESS 0	/* X server default */
 #define SERVER_FAILURE 1	/* X server default */
@@ -138,6 +145,18 @@
 #define GDM_KEY_TIMED_LOGIN "daemon/TimedLogin="
 #define GDM_KEY_TIMED_LOGIN_DELAY "daemon/TimedLoginDelay=30"
 
+#define GDM_KEY_STANDARD_XSERVER "daemon/StandardXServer=/usr/bin/X11/X"
+#define GDM_KEY_FLEXIBLE_XSERVERS "daemon/FlexibleXServers=5"
+#define GDM_KEY_XNEST "daemon/Xnest=/usr/bin/X11/Xnest -name Xnest"
+
+/* Per server definitions */
+#define GDM_KEY_SERVER_NAME "name=Standard server"
+#define GDM_KEY_SERVER_COMMAND "command=/usr/bin/X11/X"
+/* runnable as flexi server */
+#define GDM_KEY_SERVER_FLEXIBLE "flexible=true"
+/* choosable from the login screen */
+#define GDM_KEY_SERVER_CHOOSABLE "choosable=true"
+
 #define GDM_KEY_ALLOWROOT "security/AllowRoot=true"
 #define GDM_KEY_ALLOWREMOTEROOT "security/AllowRemoteRoot=true"
 #define GDM_KEY_ALLOWREMOTEAUTOLOGIN "security/AllowRemoteAutoLogin=false"
@@ -210,6 +229,12 @@
 #define GDM_SESSION_FAILSAFE_XTERM "GDM_Failsafe.XTERM"
 #define GDM_SESSION_GNOME_CHOOSER "Gnome Chooser"
 
+#define GDM_STANDARD "Standard"
+
+#ifndef TYPEDEF_GDM_CONNECTION
+#define TYPEDEF_GDM_CONNECTION
+typedef struct _GdmConnection GdmConnection;
+#endif  /* TYPEDEF_GDM_CONNECTION */
 
 typedef struct _GdmDisplay GdmDisplay;
 
@@ -239,6 +264,14 @@ struct _GdmDisplay {
     pid_t chooserpid;
     time_t acctime;
 
+#ifdef __linux__
+    int vt;
+#endif
+
+    gboolean busy_display;
+
+    gboolean console;
+
     time_t last_start_time;
     gint retry_count;
 
@@ -250,6 +283,7 @@ struct _GdmDisplay {
     gboolean disabled;
 
     gboolean logged_in; /* TRUE if someone is logged in */
+    char *login;
 
     gboolean timed_login_ok;
 
@@ -257,8 +291,21 @@ struct _GdmDisplay {
     int screeny;
     int screenwidth; /* Note 0 means use the gdk size */
     int screenheight;
+
+    /* Flexi stuff */
+    char *xnest_disp;
+    char *xnest_auth_file;
+    GdmConnection *socket_conn;
 };
 
+typedef struct _GdmXServer GdmXServer;
+struct _GdmXServer {
+	char *id;
+	char *name;
+	char *command;
+	gboolean flexible;
+	gboolean choosable; /* not implemented yet */
+};
 
 typedef struct _GdmIndirectDisplay GdmIndirectDisplay;
 struct _GdmIndirectDisplay {
@@ -309,24 +356,134 @@ guint   g_signal_add_full       (gint           priority,
 				 GDestroyNotify destroy);
 void    g_signal_notify         (gint8          signal);
 
+/* If id == NULL, then get the first X server */
+GdmXServer *	gdm_find_x_server	(const char *id);
+
 
 void gdm_run (void);
 void gdm_quit (void);
 
-/* primitive protocol for controlling the daemon from the chooser
+/* primitive protocol for controlling the daemon from slave
  * or gdmconfig or whatnot */
 
 /* The ones that pass a <slave pid> must be from a valid slave, and
  * the slave will be sent a SIGUSR2 */
+/* The fifo protocol, used only by gdm internally */
 #define GDM_SOP_CHOSEN       "CHOSEN" /* <indirect id> <ip addr> */
 #define GDM_SOP_XPID         "XPID" /* <slave pid> <xpid> */
 #define GDM_SOP_SESSPID      "SESSPID" /* <slave pid> <sesspid> */
 #define GDM_SOP_GREETPID     "GREETPID" /* <slave pid> <greetpid> */
 #define GDM_SOP_CHOOSERPID   "CHOOSERPID" /* <slave pid> <chooserpid> */
 #define GDM_SOP_LOGGED_IN    "LOGGED_IN" /* <slave pid> <logged_in as int> */
+#define GDM_SOP_LOGIN        "LOGIN" /* <slave pid> <username> */
+#define GDM_SOP_COOKIE       "COOKIE" /* <slave pid> <cookie> */
+#define GDM_SOP_DISP_NUM     "DISP_NUM" /* <slave pid> <display as int> */
+/* For linux only currently */
+#define GDM_SOP_VT_NUM       "VT_NUM" /* <slave pid> <vt as int> */
+#define GDM_SOP_FLEXI_ERR    "FLEXI_ERR" /* <slave pid> <error num> */
+	/* 3 = X failed */
+	/* 4 = X too busy */
+	/* 5 = Xnest can't connect */
+#define GDM_SOP_FLEXI_OK     "FLEXI_OK" /* <slave pid> */
 #define GDM_SOP_SOFT_RESTART "SOFT_RESTART" /* no arguments */
 
-void gdm_fifo_close (void);
+#define GDM_SUP_SOCKET "/tmp/.gdm_socket"
+
+/*
+ * The user socket protocol.  Each command is given on a separate line
+ *
+ * A user should first send a VERSION\n after connecting and only do
+ * anything else if gdm responds with the correct response.  The version
+ * is the gdm version and not a "protocol" revision, so you can't check
+ * against a single version but check if the version is higher then some
+ * value.
+ */
+/* The user protocol, using /tmp/.gdm_socket */
+
+#define GDM_SUP_VERSION "VERSION" /* no arguments */
+/* VERSION: Query version
+ * Supported since: 2.2.4.0
+ * Arguments:  None
+ * Answers:
+ *   GDM <gdm version>
+ */
+#define GDM_SUP_AUTH_LOCAL "AUTH_LOCAL" /* <xauth cookie> */
+/* AUTH_LOCAL: Setup this connection as authenticated for FLEXI_SERVER
+ *             Because all full blown (non-Xnest) servers can be started
+ *             only from users logged in locally, and here gdm assumes
+ *             only users logged in from gdm.  They must pass the xauth
+ *             MIT-MAGIC-COOKIE-1 that they were passed before the
+ *             connection is authenticated.
+ * Supported since: 2.2.4.0
+ * Arguments:  <xauth cookie>
+ *   <xauth cookie> is in hex form with no 0x prefix
+ * Answers:
+ *   OK
+ *   ERROR <err number> <english error description>
+ *      0 = Not implemented
+ *      100 = Not authenticated
+ *      999 = Unknown error
+ */
+#define GDM_SUP_FLEXI_XSERVER "FLEXI_XSERVER" /* <xserver type> */
+/* FLEXI_XSERVER: Start a new X flexible server
+ *   Only supported on connection that passed AUTH_LOCAL
+ * Supported since: 2.2.4.0
+ * Arguments:  <xserver type>
+ *   If no arguments, starts the standard x server
+ * Answers:
+ *   OK <display>
+ *   ERROR <err number> <english error description>
+ *      0 = Not implemented
+ *      1 = No more flexi servers
+ *      2 = Startup errors
+ *      3 = X failed
+ *      4 = X too busy
+ *      6 = No server binary
+ *      100 = Not authenticated
+ *      999 = Unknown error
+ */
+#define GDM_SUP_FLEXI_XNEST  "FLEXI_XNEST" /* <display> <xauth file> */
+/* FLEXI_XNEXT: Start a new flexible Xnest server
+ * Supported since: 2.2.4.0
+ * Arguments:  <display> <xauth file>
+ * Answers:
+ *   OK <display>
+ *   ERROR <err number> <english error description>
+ *      0 = Not implemented
+ *      1 = No more flexi servers
+ *      2 = Startup errors
+ *      3 = X failed
+ *      4 = X too busy
+ *      5 = Xnest can't connect
+ *      6 = No server binary
+ *      999 = Unknown error
+ */
+#define GDM_SUP_CONSOLE_SERVERS  "CONSOLE_SERVERS" /* None */
+/* CONSOLE_SERVERS: List all console servers, useful for linux mostly
+ *  Doesn't list xdmcp and xnest non-console servers
+ * Supported since: 2.2.4.0
+ * Arguments:  None
+ * Answers:
+ *   OK <server>;<server>;...
+ *
+ *   <server> is <display>,<logged in user>,<vt or xnest display>
+ *
+ *   <logged in user> can be empty in case no one logged in yet,
+ *   and <vt> can be -1 if it's not known or not supported (on non-linux
+ *   for example).  If the display is an xnest display and is a console one
+ *   (that is, it is an xnest inside another console display) it is listed
+ *   and instead of vt, it lists the parent display in standard form.
+ */
+#define GDM_SUP_CLOSE        "CLOSE" /* no arguments */
+/* CLOSE Answers: None
+ * Supported since: 2.2.4.0
+ */
+
+/* User flags for the SUP protocol */
+enum {
+	GDM_SUP_FLAG_AUTHENTICATED = 0x1 /* authenticated as a local user,
+					  * from a local display we started */
+};
 
 #endif /* GDM_H */
 
