@@ -39,8 +39,9 @@
 #include "xdmcp.h"
 #include "display.h"
 #include "auth.h"
+#include "slave.h"
 
-static const gchar RCSid[]="$Id: server.c,v 1.28 2001/08/02 16:50:29 hp Exp $";
+static const gchar RCSid[]="$Id: server.c,v 1.35 2001/08/26 01:56:37 jirka Exp $";
 
 #define SERVER_WAIT_ALARM 10
 
@@ -55,8 +56,10 @@ static void gdm_server_child_handler (gint);
 extern gchar *GdmDisplayInit;
 extern gchar *GdmServAuthDir;
 extern gchar *GdmLogDir;
+extern gchar *GdmStandardXServer;
 extern gboolean GdmXdmcp;
 extern sigset_t sysmask;
+extern gint high_display_num;
 
 /* Global vars */
 static GdmDisplay *d = NULL;
@@ -124,58 +127,69 @@ gdm_server_stop (GdmDisplay *disp)
     unlink (disp->authfile);
 }
 
-static void
-busy_inform_user (GdmDisplay *disp)
+static gboolean
+busy_ask_user (GdmDisplay *disp)
 {
     /* if we have "open" we can talk to the user */
-    if (access ("/usr/bin/open", X_OK) == 0) {
-	    char *dialog; /* do we have dialog?*/
+    if (access (EXPANDED_SBINDIR "/gdmopen", X_OK) == 0) {
 	    char *error = g_strdup_printf
 		    (_("There already appears to be an X server "
-		       "running on display %s.  Please quit this "
-		       "server and then press Enter.%s"),
+		       "running on display %s.  Should I try another "
+		       "display number?  If you answer no, I will "
+		       "attempt to start the server on %s again.%s"),
+		     disp->name,
 		     disp->name,
 #ifdef __linux__
-		     _("  You can change consoles by pressing Ctrl-Alt "
+		     _("  (You can change consoles by pressing Ctrl-Alt "
 		       "plus a function key, such as Ctrl-Alt-F7 to go "
 		       "to console 7.  X servers usually run on consoles "
-		       "7 and higher.")
+		       "7 and higher.)")
 #else /* ! __linux__ */
 		     /* no info for non linux users */
 		     ""
 #endif /* __linux__ */
 		     );
-	    dialog = gnome_is_program_in_path ("dialog");
-	    if (dialog == NULL)
-		    dialog = gnome_is_program_in_path ("gdialog");
-	    if (dialog != NULL) {
-		    char *command = 
-			    g_strdup_printf
-			    ("/usr/bin/open -s -w -- /bin/sh -c 'clear ; "
-			     "%s --msgbox \"%s\" 10 70 ; clear'",
-			     dialog,
-			     error);
-		    system (command);
-		    g_free (command);
-		    g_free (dialog);
-	    } else {
-		    char *command = 
-			    g_strdup_printf
-			    ("/usr/bin/open -s -w -- /bin/sh -c 'clear ; "
-			     "echo \"%s\" 10 70 ; read ; clear'",
-			     error);
-		    system (command);
-		    g_free (command);
-	    }
+	    gboolean ret = TRUE;
+	    /* default ret to TRUE */
+	    if ( ! gdm_text_yesno_dialog (error, &ret))
+		    ret = TRUE;
 	    g_free (error);
+	    return ret;
     } else {
-	    /* If we can't ask, sleep 30 seconds and try again */
-	    gdm_info (_("Sleeping 30 seconds before retrying display %s"),
-		      disp->name);
-	    sleep (30);
+	    /* Well we'll just try another display number */
+	    return TRUE;
     }
 }
 
+static gboolean
+display_xnest_no_connect (GdmDisplay *disp)
+{
+	char *logname = g_strconcat (GdmLogDir, "/", d->name, ".log", NULL);
+	FILE *fp;
+	char buf[256];
+
+	fp = fopen (logname, "r");
+	g_free (logname);
+
+	if (fp == NULL)
+		return FALSE;
+
+	while (fgets (buf, sizeof (buf), fp) != NULL) {
+		/* Note: this is probably XFree86 specific, and perhaps even
+		 * version 3 specific (I don't have xfree v4 to test this),
+		 * of course additions are welcome to make this more robust */
+		if (strstr (buf, "Unable to open display \"") == buf) {
+			gdm_error (_("Display '%s' cannot be opened by Xnest"),
+				   ve_sure_string (disp->xnest_disp));
+			fclose (fp);
+			return TRUE;
+		}
+	}
+
+	fclose (fp);
+	return FALSE;
+}
+
 static gboolean
 display_busy (GdmDisplay *disp)
 {
@@ -194,7 +208,7 @@ display_busy (GdmDisplay *disp)
 		 * version 3 specific (I don't have xfree v4 to test this),
 		 * of course additions are welcome to make this more robust */
 		if (strstr (buf, "Server is already active for display")
-		    != NULL) {
+		    == buf) {
 			gdm_error (_("Display %s is busy, there is another "
 				     "X server already running"),
 				   disp->name);
@@ -207,6 +221,35 @@ display_busy (GdmDisplay *disp)
 	return FALSE;
 }
 
+#ifdef __linux__
+static int
+display_vt (GdmDisplay *disp)
+{
+	char *logname = g_strconcat (GdmLogDir, "/", d->name, ".log", NULL);
+	FILE *fp;
+	char buf[256];
+
+	fp = fopen (logname, "r");
+	g_free (logname);
+
+	if (fp == NULL)
+		return FALSE;
+
+	while (fgets (buf, sizeof (buf), fp) != NULL) {
+		int vt;
+		/* Note: this is probably XFree86 specific, and perhaps even
+		 * version 3 specific (I don't have xfree v4 to test this),
+		 * of course additions are welcome to make this more robust */
+		if (sscanf (buf, "(using VT number %d)", &vt) == 1) {
+			fclose (fp);
+			return vt;
+		}
+	}
+	fclose (fp);
+	return -1;
+}
+#endif
+
 /**
  * gdm_server_start:
  * @disp: Pointer to a GdmDisplay structure
@@ -215,26 +258,42 @@ display_busy (GdmDisplay *disp)
  */
 
 gboolean
-gdm_server_start (GdmDisplay *disp)
+gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
+		  int min_flexi_disp, int flexi_retries)
 {
     struct sigaction usr1, chld, alrm;
     struct sigaction old_usr1, old_chld, old_alrm;
     sigset_t mask, oldmask;
-    gboolean retvalue;
-    
-    if (!disp)
-	return FALSE;
+    int flexi_disp = 20;
     
+    if (disp == NULL)
+	    return FALSE;
+
     d = disp;
 
     /* if an X server exists, wipe it */
     gdm_server_stop (d);
 
+    if (SERVER_IS_FLEXI (d) ||
+	treat_as_flexi) {
+	    flexi_disp = gdm_get_free_display
+		    (MAX (high_display_num + 1, min_flexi_disp) /* start */,
+		     0 /* server uid */);
+
+	    g_free (d->name);
+	    d->name = g_strdup_printf (":%d", flexi_disp);
+	    d->dispnum = flexi_disp;
+
+	    gdm_slave_send_num (GDM_SOP_DISP_NUM, flexi_disp);
+    }
+
+
     gdm_debug ("gdm_server_start: %s", d->name);
 
     /* Create new cookie */
     if ( ! gdm_auth_secure_display (d))
 	    return FALSE;
+    gdm_slave_send_string (GDM_SOP_COOKIE, d->cookie);
     ve_setenv ("DISPLAY", d->name, TRUE);
 
     /* Catch USR1 from X server */
@@ -289,8 +348,6 @@ gdm_server_start (GdmDisplay *disp)
     /* fork X server process */
     gdm_server_spawn (d);
 
-    retvalue = FALSE;
-
     /* Wait for X server to send ready signal */
     if (d->servstat == SERVER_STARTED)
 	    gdm_run ();
@@ -311,9 +368,25 @@ gdm_server_start (GdmDisplay *disp)
     case SERVER_RUNNING:
 	    gdm_debug ("gdm_server_start: Completed %s!", d->name);
 
-	    retvalue = TRUE;
-	    goto spawn_done;
+	    sigprocmask (SIG_SETMASK, &oldmask, NULL);
 
+	    /* restore default handlers */
+	    sigaction (SIGUSR1, &old_usr1, NULL);
+	    sigaction (SIGCHLD, &old_chld, NULL);
+	    sigaction (SIGALRM, &old_alrm, NULL);
+
+	    if (SERVER_IS_FLEXI (d))
+		    gdm_slave_send_num (GDM_SOP_FLEXI_OK, 0 /* bogus */);
+#ifdef __linux__
+	    if (d->type == TYPE_LOCAL ||
+		d->type == TYPE_FLEXI) {
+		    d->vt = display_vt (d);
+		    if (d->vt >= 0)
+			    gdm_slave_send_num (GDM_SOP_VT_NUM, d->vt);
+	    }
+#endif
+
+	    return TRUE;
     default:
 	    break;
     }
@@ -325,19 +398,6 @@ gdm_server_start (GdmDisplay *disp)
 	    d->servpid = 0;
     }
 
-    /* if this was a busy fail, that is, there is already
-     * a server on that display, we'll display an error and after
-     * this we'll exit with DISPLAY_REMANAGE to try again if the
-     * user wants to, or abort this display */
-    if (display_busy (disp)) {
-	    busy_inform_user (disp);
-	    _exit (DISPLAY_REMANAGE);
-    }
-
-    _exit (DISPLAY_XFAILED);
-
-spawn_done:
-
     sigprocmask (SIG_SETMASK, &oldmask, NULL);
 
     /* restore default handlers */
@@ -345,9 +405,74 @@ spawn_done:
     sigaction (SIGCHLD, &old_chld, NULL);
     sigaction (SIGALRM, &old_alrm, NULL);
 
-    return retvalue;
+    if (disp->type == TYPE_FLEXI_XNEST &&
+	display_xnest_no_connect (disp)) {
+	    gdm_slave_send_num (GDM_SOP_FLEXI_ERR,
+				5 /* Xnest can't connect */);
+	    _exit (DISPLAY_REMANAGE);
+    }
+
+    /* if this was a busy fail, that is, there is already
+     * a server on that display, we'll display an error and after
+     * this we'll exit with DISPLAY_REMANAGE to try again if the
+     * user wants to, or abort this display */
+    if (display_busy (disp)) {
+	    if (SERVER_IS_FLEXI (disp) ||
+		treat_as_flexi) {
+		    /* for flexi displays, try again a few times with different
+		     * display numbers */
+		    if (flexi_retries <= 0) {
+			    /* Send X too busy */
+			    gdm_error (_("%s: Cannot find a free "
+					 "display number"),
+				       "gdm_server_start");
+			    if (SERVER_IS_FLEXI (disp)) {
+				    gdm_slave_send_num (GDM_SOP_FLEXI_ERR,
+							4 /* X too busy */);
+			    }
+			    /* eki eki */
+			    _exit (DISPLAY_REMANAGE);
+		    }
+		    return gdm_server_start (d, treat_as_flexi,
+					     flexi_disp + 1,
+					     flexi_retries - 1);
+	    } else {
+		    if (busy_ask_user (disp)) {
+			    gdm_error (_("%s: Display %s busy.  Trying "
+					 "another display number."),
+				       "gdm_server_start",
+				       d->name);
+			    d->busy_display = TRUE;
+			    return gdm_server_start (d,
+						     TRUE /* treat as flexi */,
+						     high_display_num + 1,
+						     flexi_retries - 1);
+		    }
+		    _exit (DISPLAY_REMANAGE);
+	    }
+    }
+
+    _exit (DISPLAY_XFAILED);
+
+    return FALSE;
 }
 
+/* Do things that require checking the log,
+ * we really do need to get called a bit later, after all init is done
+ * as things aren't written to disk before that */
+void
+gdm_server_checklog (GdmDisplay *disp)
+{
+#ifdef __linux__
+	    if (d->vt < 0 &&
+		(d->type == TYPE_LOCAL ||
+		 d->type == TYPE_FLEXI)) {
+		    d->vt = display_vt (d);
+		    if (d->vt >= 0)
+			    gdm_slave_send_num (GDM_SOP_VT_NUM, d->vt);
+	    }
+#endif
+}
 
 /**
  * gdm_server_spawn:
@@ -365,6 +490,8 @@ gdm_server_spawn (GdmDisplay *d)
     gchar **argv = NULL;
     int logfd;
     int len, i;
+    const char *command;
+    char *bin;
 
     if (d == NULL ||
 	ve_string_empty (d->command)) {
@@ -378,7 +505,7 @@ gdm_server_spawn (GdmDisplay *d)
 	    if (kill (d->servpid, SIGTERM) == 0)
 		    waitpid (d->servpid, NULL, 0);
     }
-
+    
     /* Fork into two processes. Parent remains the gdm process. Child
      * becomes the X server. */
     
@@ -391,6 +518,10 @@ gdm_server_spawn (GdmDisplay *d)
 	if (GdmXdmcp)
 	    gdm_xdmcp_close();
 
+	/* close things */
+	for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
+		close(i);
+
         /* Log all output from spawned programs to a file */
 	logfd = open (g_strconcat (GdmLogDir, "/", d->name, ".log", NULL),
 		      O_CREAT|O_TRUNC|O_WRONLY, 0666);
@@ -444,7 +575,50 @@ gdm_server_spawn (GdmDisplay *d)
 	sigaddset (&mask, SIGTERM);
 	sigprocmask (SIG_UNBLOCK, &mask, NULL);
 
-	argv = ve_split (d->command);
+	if (d->type == TYPE_FLEXI_XNEST) {
+		ve_setenv ("DISPLAY", d->xnest_disp, TRUE);
+		if (d->xnest_auth_file != NULL)
+			ve_setenv ("XAUTHORITY", d->xnest_auth_file, TRUE);
+		else
+			ve_unsetenv ("XAUTHORITY");
+	}
+
+	bin = ve_first_word (d->command);
+	if (bin == NULL) {
+		gdm_error (_("Invalid server command '%s'"), d->command);
+		argv = ve_split (GdmStandardXServer);
+		command = GdmStandardXServer;
+	} else if (bin[0] != '/') {
+		GdmXServer *svr = gdm_find_x_server (bin);
+		if (svr == NULL) {
+			gdm_error (_("Server name '%s' not found, "
+				     "using standard server"), bin);
+			argv = ve_split (GdmStandardXServer);
+			command = GdmStandardXServer;
+		} else {
+			char **svr_command =
+				ve_split (ve_sure_string (svr->command));
+			argv = ve_split (d->command);
+			if (argv[0] == NULL || argv[1] == NULL) {
+				g_strfreev (argv);
+				argv = svr_command;
+			} else {
+				char **old_argv = argv;
+				argv = ve_vector_merge (svr_command,
+							&old_argv[1]);
+				g_strfreev (svr_command);
+				g_strfreev (old_argv);
+			}
+			/* this is only for information only,
+			 * so doesn't include whole command line */
+			command = svr->command;
+		}
+	} else {
+		command = d->command;
+		argv = ve_split (d->command);
+	}
+	g_free (bin);
+
 	for (len = 0; argv != NULL && argv[len] != NULL; len++)
 		;
 
@@ -473,7 +647,7 @@ gdm_server_spawn (GdmDisplay *d)
 	setpgid (0, 0);
 	execv (argv[0], argv);
 	
-	gdm_error (_("gdm_server_spawn: Xserver not found: %s"), d->command);
+	gdm_error (_("gdm_server_spawn: Xserver not found: %s"), command);
 	
 	_exit (SERVER_ABORT);
 	
@@ -578,16 +752,22 @@ gdm_server_alloc (gint id, const gchar *command)
     d->sesspid = 0;
     d->slavepid = 0;
     d->type = TYPE_LOCAL;
+    d->console = TRUE;
     d->sessionid = 0;
     d->acctime = 0;
     d->dsp = NULL;
     d->screenx = 0; /* xinerama offset */
     d->screeny = 0;
 
+#ifdef __linux__
+    d->vt = -1;
+#endif
+
     d->last_start_time = 0;
     d->retry_count = 0;
     d->disabled = FALSE;
     d->sleep_before_run = 0;
+    d->login = NULL;
 
     d->timed_login_ok = FALSE;
     
