@@ -57,7 +57,7 @@
 #include "errorgui.h"
 
 
-static const gchar RCSid[]="$Id: slave.c,v 1.83 2001/08/10 19:14:18 jirka Exp $";
+static const gchar RCSid[]="$Id: slave.c,v 1.90 2001/08/26 01:56:37 jirka Exp $";
 
 /* Some per slave globals */
 static GdmDisplay *d;
@@ -74,6 +74,7 @@ static gchar *ParsedTimedLogin = NULL;
 
 extern gboolean gdm_first_login;
 extern gboolean gdm_emergency_server;
+extern pid_t extra_process;
 
 /* Configuration option variables */
 extern gchar *GdmUser;
@@ -112,6 +113,7 @@ extern gboolean GdmAllowRoot;
 extern sigset_t sysmask;
 extern gchar *GdmGlobalFaceDir;
 extern gboolean GdmBrowser;
+extern gboolean GdmDebug;
 
 
 /* Local prototypes */
@@ -133,8 +135,6 @@ static void     gdm_child_exit (gint status, const gchar *format, ...) G_GNUC_PR
 static gint     gdm_slave_exec_script (GdmDisplay *d, const gchar *dir,
 				       const char *login, struct passwd *pwent);
 static gchar *  gdm_parse_enriched_login (const gchar *s, GdmDisplay *display);
-static void	gdm_send_logged_in (gboolean logged_in);
-static void	gdm_send_pid (const char *opcode, pid_t pid);
 
 
 /* Yay thread unsafety */
@@ -160,7 +160,7 @@ gdm_slave_start (GdmDisplay *display)
 		return;
 
 	gdm_debug ("gdm_slave_start: Starting slave process for %s", display->name);
-	if (display->type != TYPE_LOCAL &&
+	if (display->type == TYPE_XDMCP &&
 	    GdmPingInterval > 0) {
 		/* Handle a ALRM signals from our ping alarms */
 		alrm.sa_handler = gdm_slave_alrm_handler;
@@ -207,7 +207,7 @@ gdm_slave_start (GdmDisplay *display)
 	sigdelset (&mask, SIGTERM);
 	sigdelset (&mask, SIGCHLD);
 	sigdelset (&mask, SIGUSR2);
-	if (display->type != TYPE_LOCAL &&
+	if (display->type == TYPE_XDMCP &&
 	    GdmPingInterval > 0) {
 		sigdelset (&mask, SIGALRM);
 	}
@@ -222,6 +222,7 @@ gdm_slave_start (GdmDisplay *display)
 		gdm_debug ("gdm_slave_start: Loop Thingie");
 		gdm_slave_run (display);
 
+		/* remote and flexi only run once */
 		if (display->type != TYPE_LOCAL)
 			break;
 
@@ -246,12 +247,13 @@ gdm_slave_start (GdmDisplay *display)
 			/* Whack the server if we want to restart it next time
 			 * we run gdm_slave_run */
 			gdm_server_stop (display);
-			gdm_send_pid (GDM_SOP_XPID, 0);
+			gdm_slave_send_num (GDM_SOP_XPID, 0);
 		} else {
-			/* OK about to start again so redo our cookies and reinit
+			/* OK about to start again so rebake our cookies and reinit
 			 * the server */
 			if ( ! gdm_auth_secure_display (d))
 				break;
+			gdm_slave_send_string (GDM_SOP_COOKIE, d->cookie);
 
 			gdm_server_reinit (d);
 		}
@@ -353,10 +355,13 @@ gdm_slave_run (GdmDisplay *display)
 
     /* if this is local display start a server if one doesn't
      * exist */
-    if (d->type == TYPE_LOCAL &&
+    if (SERVER_IS_LOCAL (d) &&
 	d->servpid <= 0) {
-	    gdm_server_start (d);
-	    gdm_send_pid (GDM_SOP_XPID, d->servpid);
+	    gdm_server_start (d,
+			      FALSE /* treat_as_flexi */,
+			      20 /* min_flexi_disp */,
+			      5 /* flexi_retries */);
+	    gdm_slave_send_num (GDM_SOP_XPID, d->servpid);
     }
     
     ve_setenv ("XAUTHORITY", d->authfile, TRUE);
@@ -375,7 +380,7 @@ gdm_slave_run (GdmDisplay *display)
 
     /* if local then the the server should be ready for openning, so
      * don't try so long before killing it and trying again */
-    if (d->type == TYPE_LOCAL)
+    if (SERVER_IS_LOCAL (d))
 	    maxtries = 2;
     else
 	    maxtries = 10;
@@ -391,8 +396,8 @@ gdm_slave_run (GdmDisplay *display)
 	}
     }
 
-    /* something may have gone wrong, try remanaging, if local, the toplevel
-     * loop of death will handle us */ 
+    /* something may have gone wrong, try remanaging, if local (non-flexi),
+     * the toplevel loop of death will handle us */ 
     if (d->dsp == NULL) {
 	    gdm_server_stop (d);
 	    if (d->type == TYPE_LOCAL)
@@ -414,7 +419,7 @@ gdm_slave_run (GdmDisplay *display)
     do_xfailed_on_xio_error = FALSE;
 
     /* If XDMCP setup pinging */
-    if (d->type != TYPE_LOCAL &&
+    if (d->type == TYPE_XDMCP &&
 	GdmPingInterval > 0) {
 	    alarm (GdmPingInterval * 60);
     }
@@ -422,22 +427,29 @@ gdm_slave_run (GdmDisplay *display)
     /* checkout xinerama */
     gdm_screen_init (d);
 
+    /* check log stuff for the server, this is done here
+     * because it's really a race */
+    if (SERVER_IS_LOCAL (d))
+	    gdm_server_checklog (d);
+
     if (d->use_chooser) {
 	    /* this usually doesn't return */
 	    gdm_slave_chooser ();  /* Run the chooser */
-    } else if ((d->type == TYPE_LOCAL || GdmAllowRemoteAutoLogin) &&
+    } else if ((d->console || GdmAllowRemoteAutoLogin) &&
 	       gdm_first_login &&
 	       ! ve_string_empty (ParsedAutomaticLogin) &&
 	       strcmp (ParsedAutomaticLogin, "root") != 0) {
 	    gdm_first_login = FALSE;
 
-	    gdm_send_logged_in (TRUE);
+	    gdm_slave_send_num (GDM_SOP_LOGGED_IN, TRUE);
+	    gdm_slave_send_string (GDM_SOP_LOGIN, ParsedAutomaticLogin);
 
 	    setup_automatic_session (d, ParsedAutomaticLogin);
 
 	    gdm_slave_session_start();
 
-	    gdm_send_logged_in (FALSE);
+	    gdm_slave_send_num (GDM_SOP_LOGGED_IN, FALSE);
+	    gdm_slave_send_string (GDM_SOP_LOGIN, "");
 
 	    gdm_debug ("gdm_slave_run: Automatic login done");
     } else {
@@ -446,16 +458,21 @@ gdm_slave_run (GdmDisplay *display)
 	    gdm_slave_greeter ();  /* Start the greeter */
 	    gdm_slave_wait_for_login (); /* wait for a password */
 
-	    gdm_send_logged_in (TRUE);
+	    gdm_slave_send_num (GDM_SOP_LOGGED_IN, TRUE);
 
 	    if (do_timed_login) {
 		    /* timed out into a timed login */
 		    do_timed_login = FALSE;
 		    setup_automatic_session (d, ParsedTimedLogin);
+		    gdm_slave_send_string (GDM_SOP_LOGIN,
+					   ParsedTimedLogin);
+	    } else {
+		    gdm_slave_send_string (GDM_SOP_LOGIN, login);
 	    }
 	    gdm_slave_session_start ();
 
-	    gdm_send_logged_in (FALSE);
+	    gdm_slave_send_num (GDM_SOP_LOGGED_IN, FALSE);
+	    gdm_slave_send_string (GDM_SOP_LOGIN, "");
     }
 }
 
@@ -481,7 +498,7 @@ gdm_slave_whack_greeter (void)
 		waitpid (d->greetpid, 0, 0); 
 	d->greetpid = 0;
 
-	gdm_send_pid (GDM_SOP_GREETPID, 0);
+	gdm_slave_send_num (GDM_SOP_GREETPID, 0);
 
 	sigprocmask (SIG_SETMASK, &omask, NULL);
 }
@@ -494,6 +511,7 @@ focus_first_x_window (const char *class_res_name)
 	pid_t pid;
 	Display *disp;
 	XWindowAttributes attribs = { 0, };
+	int i;
 
 	pid = fork ();
 	if (pid < 0) {
@@ -505,6 +523,15 @@ focus_first_x_window (const char *class_res_name)
 		return;
 	}
 
+	for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
+	    close(i);
+
+	/* No error checking here - if it's messed the best response
+         * is to ignore & try to continue */
+	open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
+	open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
+	open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+
 	disp = XOpenDisplay (d->name);
 	if (disp == NULL) {
 		gdm_error (_("focus_first_x_window: cannot open display %s"),
@@ -591,9 +618,9 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 
 		/* No error checking here - if it's messed the best response
 		 * is to ignore & try to continue */
-		open("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-		open("/dev/null", O_RDWR); /* open stdout - fd 1 */
-		open("/dev/null", O_RDWR); /* open stderr - fd 2 */
+		open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
+		open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
+		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
 
 		/* exec the configurator */
 		argv = ve_split (GdmConfigurator);
@@ -651,7 +678,7 @@ gdm_slave_wait_for_login (void)
 		gdm_debug ("gdm_slave_wait_for_login: In loop");
 		login = gdm_verify_user (NULL /* username*/,
 					 d->name,
-					 d->type == TYPE_LOCAL);
+					 d->console);
 		gdm_debug ("gdm_slave_wait_for_login: end verify for '%s'",
 			   ve_sure_string (login));
 		/* Complex, make sure to always handle the do_configurator
@@ -677,7 +704,7 @@ gdm_slave_wait_for_login (void)
 			gdm_slave_greeter_ctl_no_ret (GDM_SETLOGIN, "root");
 			login = gdm_verify_user ("root",
 						 d->name,
-						 d->type == TYPE_LOCAL);
+						 d->console);
 			GdmAllowRoot = oldAllowRoot;
 
 			/* the wanker can't remember his password */
@@ -719,7 +746,9 @@ gdm_slave_wait_for_login (void)
 				continue;
 			}
 
-			gdm_send_logged_in (TRUE);
+			gdm_slave_send_num (GDM_SOP_LOGGED_IN, TRUE);
+			/* Note: nobody really logged in */
+			gdm_slave_send_string (GDM_SOP_LOGIN, "");
 
 			/* disable the login screen, we don't want people to
 			 * log in in the meantime */
@@ -732,7 +761,7 @@ gdm_slave_wait_for_login (void)
 
 			gdm_verify_cleanup ();
 
-			gdm_send_logged_in (FALSE);
+			gdm_slave_send_num (GDM_SOP_LOGGED_IN, FALSE);
 
 			gdm_slave_greeter_ctl_no_ret (GDM_ENABLE, "");
 			gdm_slave_greeter_ctl_no_ret (GDM_RESETOK, "");
@@ -972,6 +1001,7 @@ gdm_slave_greeter (void)
     gint pipe1[2], pipe2[2];  
     gchar **argv;
     struct passwd *pwent;
+    int i;
     
     gdm_debug ("gdm_slave_greeter: Running greeter on %s", d->name);
     
@@ -1006,6 +1036,11 @@ gdm_slave_greeter (void)
 	
 	if (pipe2[1] != STDOUT_FILENO) 
 	    dup2 (pipe2[1], STDOUT_FILENO);
+
+	for (i = 2; i < sysconf (_SC_OPEN_MAX); i++)
+	    close(i);
+
+	open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
 	
 	if (setgid (GdmGroupId) < 0) 
 	    gdm_child_exit (DISPLAY_ABORT, _("gdm_slave_greeter: Couldn't set groupid to %d"), GdmGroupId);
@@ -1044,7 +1079,7 @@ gdm_slave_greeter (void)
 	/* Note that this is just informative, the slave will not listen to
 	 * the greeter even if it does something it shouldn't on a non-local
 	 * display so it's not a security risk */
-	if (d->type == TYPE_LOCAL) {
+	if (d->console) {
 		ve_setenv ("GDM_IS_LOCAL", "yes", TRUE);
 	} else {
 		ve_unsetenv ("GDM_IS_LOCAL");
@@ -1058,6 +1093,14 @@ gdm_slave_greeter (void)
 		ve_unsetenv ("GDM_TIMED_LOGIN_OK");
 	}
 
+	if (d->type == TYPE_FLEXI) {
+		ve_setenv ("GDM_FLEXI_SERVER", "yes", TRUE);
+	} else if (d->type == TYPE_FLEXI_XNEST) {
+		ve_setenv ("GDM_FLEXI_SERVER", "Xnest", TRUE);
+	} else {
+		ve_unsetenv ("GDM_FLEXI_SERVER");
+	}
+
 	if(gdm_emergency_server) {
 		gdm_error_box (d,
 			       GNOME_MESSAGE_BOX_ERROR,
@@ -1082,6 +1125,15 @@ gdm_slave_greeter (void)
 				 "configure the X server."));
 	}
 
+	if (d->busy_display) {
+		char *msg = g_strdup_printf
+			(_("The specified display number was busy, so "
+			   "this server was started on display %s."),
+			 d->name);
+		gdm_error_box (d, GNOME_MESSAGE_BOX_ERROR, msg);
+		g_free (msg);
+	}
+
 	argv = ve_split (GdmGreeter);
 	execv (argv[0], argv);
 
@@ -1137,15 +1189,15 @@ gdm_slave_greeter (void)
 	
 	gdm_debug ("gdm_slave_greeter: Greeter on pid %d", d->greetpid);
 
-	gdm_send_pid (GDM_SOP_GREETPID, d->greetpid);
+	gdm_slave_send_num (GDM_SOP_GREETPID, d->greetpid);
 
 	run_pictures (); /* Append pictures to greeter if browsing is on */
 	break;
     }
 }
 
-static void
-gdm_send_pid (const char *opcode, pid_t pid)
+void
+gdm_slave_send_num (const char *opcode, long num)
 {
 	char *msg;
 	int fd;
@@ -1153,7 +1205,7 @@ gdm_send_pid (const char *opcode, pid_t pid)
 
 	gdm_debug ("Sending %s == %ld for slave %ld",
 		   opcode,
-		   (long)pid,
+		   (long)num,
 		   (long)getpid ());
 
 	gdm_got_usr2 = FALSE;
@@ -1163,12 +1215,12 @@ gdm_send_pid (const char *opcode, pid_t pid)
 	fd = open (fifopath, O_WRONLY);
 	/* eek */
 	if (fd < 0) {
-		gdm_error (_("%s: Can't open fifo!"), "gdm_send_pid");
+		gdm_error (_("%s: Can't open fifo!"), "gdm_send_num");
 		return;
 	}
 
 	msg = g_strdup_printf ("\n%s %ld %ld\n", opcode,
-			       (long)getpid (), (long)pid);
+			       (long)getpid (), (long)num);
 
 	write (fd, msg, strlen (msg));
 
@@ -1180,17 +1232,25 @@ gdm_send_pid (const char *opcode, pid_t pid)
 		sleep (10);
 }
 
-
-static void
-gdm_send_logged_in (gboolean logged_in)
+void
+gdm_slave_send_string (const char *opcode, const char *str)
 {
 	char *msg;
 	int fd;
 	char *fifopath;
 
-	gdm_debug ("Sending logged_in == %s for slave %ld",
-		   logged_in ? "TRUE" : "FALSE",
-		   (long)getpid ());
+	/* Evil!, all this for debugging? */
+	if (GdmDebug) {
+		if (strcmp (opcode, GDM_SOP_COOKIE) == 0)
+			gdm_debug ("Sending %s == <secret> for slave %ld",
+				   opcode,
+				   (long)getpid ());
+		else
+			gdm_debug ("Sending %s == %s for slave %ld",
+				   opcode,
+				   ve_sure_string (str),
+				   (long)getpid ());
+	}
 
 	gdm_got_usr2 = FALSE;
 
@@ -1199,12 +1259,12 @@ gdm_send_logged_in (gboolean logged_in)
 	fd = open (fifopath, O_WRONLY);
 	/* eek */
 	if (fd < 0) {
-		gdm_error (_("%s: Can't open fifo!"), "gdm_send_logged_in");
+		gdm_error (_("%s: Can't open fifo!"), "gdm_slave_send_string");
 		return;
 	}
 
-	msg = g_strdup_printf ("\n%s %ld %d\n", GDM_SOP_LOGGED_IN,
-			       (long)getpid (), (int)logged_in);
+	msg = g_strdup_printf ("\n%s %ld %s\n", opcode,
+			       (long)getpid (), ve_sure_string (str));
 
 	write (fd, msg, strlen (msg));
 
@@ -1264,6 +1324,7 @@ gdm_slave_chooser (void)
 	struct passwd *pwent;
 	char buf[1024];
 	size_t bytes;
+	int i;
 
 	gdm_debug ("gdm_slave_chooser: Running chooser on %s", d->name);
 
@@ -1293,6 +1354,13 @@ gdm_slave_chooser (void)
 		if (p[1] != STDOUT_FILENO) 
 			dup2 (p[1], STDOUT_FILENO);
 
+		close (0);
+		for (i = 2; i < sysconf (_SC_OPEN_MAX); i++)
+			close(i);
+
+		open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
+		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+
 		if (setgid (GdmGroupId) < 0) 
 			gdm_child_exit (DISPLAY_ABORT, _("gdm_slave_chooser: Couldn't set groupid to %d"), GdmGroupId);
 
@@ -1351,7 +1419,7 @@ gdm_slave_chooser (void)
 		}
 
 		gdm_debug ("gdm_slave_chooser: Chooser on pid %d", d->chooserpid);
-		gdm_send_pid (GDM_SOP_CHOOSERPID, d->chooserpid);
+		gdm_slave_send_num (GDM_SOP_CHOOSERPID, d->chooserpid);
 
 		close (p[1]);
 
@@ -1360,18 +1428,18 @@ gdm_slave_chooser (void)
 		/* wait for the chooser to die */
 		waitpid (d->chooserpid, 0, 0);
 
-		gdm_send_pid (GDM_SOP_CHOOSERPID, 0);
+		gdm_slave_send_num (GDM_SOP_CHOOSERPID, 0);
 
 		bytes = read (p[0], buf, sizeof(buf)-1);
 		if (bytes > 0) {
+			close (p[0]);
+
 			if (buf[bytes-1] == '\n')
 				buf[bytes-1] ='\0';
 			else
 				buf[bytes] ='\0';
 			send_chosen_host (d, buf);
 
-			close (p[0]);
-
 			_exit (DISPLAY_CHOSEN);
 		}
 
@@ -1871,9 +1939,6 @@ gdm_slave_session_start (void)
 	ve_setenv ("LANG", language, TRUE);
 	ve_setenv ("GDM_LANG", language, TRUE);
     
-	/* setup the verify env vars */
-	gdm_verify_env_setup ();
-    
 	setpgid (0, 0);
 	
 	umask (022);
@@ -1886,6 +1951,14 @@ gdm_slave_session_start (void)
 		gdm_child_exit (DISPLAY_REMANAGE,
 				_("gdm_slave_session_start: initgroups() failed for %s. Aborting."), login);
 
+	/* setup the verify env vars, set credentials and such stuff 
+	 * and open the session */
+	if ( ! gdm_verify_open_session ())
+		gdm_child_exit (DISPLAY_REMANAGE,
+				_("%s: Could not open session for %s. "
+				  "Aborting."),
+				"gdm_slave_session_start", login);
+
 	if (setuid (pwent->pw_uid) < 0) 
 		gdm_child_exit (DISPLAY_REMANAGE,
 				_("gdm_slave_session_start: Could not become %s. Aborting."), login);
@@ -1922,21 +1995,21 @@ gdm_slave_session_start (void)
 	}
 
 	if (need_config_sync) {
-		gnome_config_sync();
+		gnome_config_sync ();
+		gnome_config_drop_all ();
 	}
 
 	for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
-	    close(i);
+	    close (i);
 
 	/* No error checking here - if it's messed the best response
          * is to ignore & try to continue */
-	open("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-	open("/dev/null", O_RDWR); /* open stdout - fd 1 */
-	open("/dev/null", O_RDWR); /* open stderr - fd 2 */
+	open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
+	open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
+	open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
 	
 	/* Restore sigmask inherited from init */
 	sigprocmask (SIG_SETMASK, &sysmask, NULL);
-	
 
 	/* If "Gnome Chooser" is still set as a session,
 	 * just change that to "Gnome", since "Gnome Chooser" is a
@@ -1945,14 +2018,6 @@ gdm_slave_session_start (void)
 		g_free (session);
 		session = g_strdup ("Gnome");
 	}
-	for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
-	    close(i);
-
-	/* No error checking here - if it's messed the best response
-         * is to ignore & try to continue */
-	open("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-	open("/dev/null", O_RDWR); /* open stdout - fd 1 */
-	open("/dev/null", O_RDWR); /* open stderr - fd 2 */
 	
 	/* Restore sigmask inherited from init */
 	sigprocmask (SIG_SETMASK, &sysmask, NULL);
@@ -2036,7 +2101,9 @@ gdm_slave_session_start (void)
 
 	/* just a stupid test, the below would fail, but this gives a better
 	 * message */
-	if (strcmp (shell, "/bin/false") == 0) {
+	if (strcmp (shell, "/sbin/nologin") == 0 ||
+	    strcmp (shell, "/bin/false") == 0 ||
+	    strcmp (shell, "/bin/true") == 0) {
 		gdm_error (_("gdm_slave_session_start: User not allowed to log in"));
 		gdm_error_box (d, GNOME_MESSAGE_BOX_ERROR,
 			       _("The system administrator has\n"
@@ -2076,7 +2143,7 @@ gdm_slave_session_start (void)
     g_free (gnome_session);
 
     sesspid = d->sesspid;
-    gdm_send_pid (GDM_SOP_SESSPID, sesspid);
+    gdm_slave_send_num (GDM_SOP_SESSPID, sesspid);
 
     /* Wait for the user's session to exit, but by this time the
      * session might have ended, so check for 0 */
@@ -2103,7 +2170,7 @@ gdm_slave_session_stop (pid_t sesspid)
     seteuid (0);
     setegid (0);
 
-    gdm_send_pid (GDM_SOP_SESSPID, 0);
+    gdm_slave_send_num (GDM_SOP_SESSPID, 0);
 
     gdm_debug ("gdm_slave_session_stop: %s on %s", local_login, d->name);
 
@@ -2169,6 +2236,12 @@ gdm_slave_term_handler (int sig)
 	sigaddset (&tmask, SIGCHLD);
 	sigprocmask (SIG_BLOCK, &tmask, &omask);  
 
+	if (extra_process > 1) {
+		/* we sigterm extra processes, and we
+		 * don't wait */
+		kill (extra_process, SIGTERM);
+		extra_process = -1;
+	}
 
 	if (d->greetpid != 0) {
 		greet = FALSE;
@@ -2434,6 +2507,7 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
     gchar *script, *defscript, *scr;
     gchar **argv;
     gint status;
+    int i;
 
     if (!d || !dir)
 	return EXIT_SUCCESS;
@@ -2448,9 +2522,19 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
     else
 	return EXIT_SUCCESS;
 
-    switch (pid = fork()) {
+    extra_process = pid = fork();
+    switch (pid) {
 	    
     case 0:
+        for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
+	    close (i);
+
+        /* No error checking here - if it's messed the best response
+         * is to ignore & try to continue */
+        open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
+        open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
+        open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+
         if (login != NULL) {
 	        ve_setenv ("LOGNAME", login, TRUE);
 	        ve_setenv ("USER", login, TRUE);
@@ -2485,6 +2569,8 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
     default:
 	waitpid (pid, &status, 0);	/* Wait for script to finish */
 
+	extra_process = -1;
+
 	if (WIFEXITED (status))
 	    return WEXITSTATUS (status);
 	else
@@ -2507,7 +2593,7 @@ gdm_slave_greeter_check_interruption (const char *msg)
 			 * it is allowed for this display (it's only allowed
 			 * for the first local display) and if it's set up
 			 * correctly */
-			if ((d->type == TYPE_LOCAL || GdmAllowRemoteAutoLogin) 
+			if ((d->console || GdmAllowRemoteAutoLogin) 
                             && d->timed_login_ok &&
 			    ! ve_string_empty (ParsedTimedLogin) &&
                             strcmp (ParsedTimedLogin, "root") != 0 &&
@@ -2517,7 +2603,7 @@ gdm_slave_greeter_check_interruption (const char *msg)
 			}
 			break;
 		case GDM_INTERRUPT_CONFIGURE:
-			if (d->type == TYPE_LOCAL &&
+			if (d->console &&
 			    GdmConfigAvailable &&
 			    GdmSystemMenu &&
 			    ! ve_string_empty (GdmConfigurator)) {
@@ -2600,49 +2686,54 @@ gdm_parse_enriched_login (const gchar *s, GdmDisplay *display)
 
     if(str->len > 0 && str->str[str->len - 1] == '|') {
       g_string_truncate(str, str->len - 1);
-      if (pipe(pipe1) < 0)
-        syslog (LOG_ERR, _("gdm_parse_enriched_login: Failed creating pipe"));
-
-      switch (pid = fork()) {
+      if (pipe (pipe1) < 0) {
+        gdm_error (_("gdm_parse_enriched_login: Failed creating pipe"));
+      } else {
+        extra_process = pid = fork();
+        switch (pid) {
 	    
-      case 0:
-	  /* The child will write the username to stdout based on the DISPLAY
-	     environment variable. */
+        case 0:
+	    /* The child will write the username to stdout based on the DISPLAY
+	       environment variable. */
 
-          close(pipe1[0]);
-          if(pipe1[1] != STDOUT_FILENO) 
-	    dup2 (pipe1[1], STDOUT_FILENO);
+            close (pipe1[0]);
+            if(pipe1[1] != STDOUT_FILENO) 
+	      dup2 (pipe1[1], STDOUT_FILENO);
 
-	  ve_setenv ("XAUTHORITY", display->authfile, TRUE);
-	  ve_setenv ("DISPLAY", display->name, TRUE);
-	  ve_setenv ("PATH", GdmRootPath, TRUE);
-	  ve_unsetenv ("MAIL");
+	    ve_setenv ("XAUTHORITY", display->authfile, TRUE);
+	    ve_setenv ("DISPLAY", display->name, TRUE);
+	    ve_setenv ("PATH", GdmRootPath, TRUE);
+	    ve_unsetenv ("MAIL");
 
-	  argv = ve_split (str->str);
-	  execv (argv[0], argv);
-	  syslog (LOG_ERR, _("gdm_parse_enriched_login: Failed executing: %s"),
-		  str->str);
-	  _exit (EXIT_SUCCESS);
+	    argv = ve_split (str->str);
+	    execv (argv[0], argv);
+	    gdm_error (_("gdm_parse_enriched_login: Failed executing: %s"),
+		    str->str);
+	    _exit (EXIT_SUCCESS);
 	    
-      case -1:
-	  syslog (LOG_ERR, _("gdm_parse_enriched_login: Can't fork script process!"));
-	  break;
+        case -1:
+	    gdm_error (_("gdm_parse_enriched_login: Can't fork script process!"));
+            close (pipe1[0]);
+            close (pipe1[1]);
+	    break;
 	
-      default:
-	  /* The parent reads username from the pipe a chunk at a time */
-          close(pipe1[1]);
-          g_string_truncate(str,0);
-          while((in_buffer_len = read(pipe1[0],in_buffer,
-				      sizeof(in_buffer)/sizeof(char)- 1)) > 0) {
-	    in_buffer[in_buffer_len] = '\000';
-            g_string_append(str,in_buffer);
-          }
+        default:
+	    /* The parent reads username from the pipe a chunk at a time */
+            close(pipe1[1]);
+            g_string_truncate(str,0);
+            while((in_buffer_len = read(pipe1[0],in_buffer,
+				        sizeof(in_buffer)/sizeof(char)- 1)) > 0) {
+	      in_buffer[in_buffer_len] = '\000';
+              g_string_append(str,in_buffer);
+            }
 
-          if(str->len > 0 && str->str[str->len - 1] == '\n')
-            g_string_truncate(str, str->len - 1);
+            if(str->len > 0 && str->str[str->len - 1] == '\n')
+              g_string_truncate(str, str->len - 1);
 
-          close(pipe1[0]);
-	  waitpid (pid, &status, 0);	/* Wait for script to finish */
+            close(pipe1[0]);
+	    waitpid (pid, &status, 0);	/* Wait for script to finish */
+	    extra_process = -1;
+        }
       }
     }
 
