@@ -95,7 +95,7 @@
 #include "choose.h"
 #include "xdmcp.h"
 
-static const gchar RCSid[]="$Id: xdmcp.c,v 1.33 2001/07/25 00:08:50 jirka Exp $";
+static const gchar RCSid[]="$Id: xdmcp.c,v 1.35 2001/08/22 23:11:54 jirka Exp $";
 gint pending = 0;
 
 #ifdef HAVE_LIBXDMCP
@@ -419,7 +419,7 @@ gdm_xdmcp_decode_packet (GIOChannel *source, GIOCondition cond, gpointer data)
 	"MANAGED_FORWARD"
     };
     
-    if (!XdmcpFill (xdmcpfd, &buf, &clnt_sa, &sa_len)) {
+    if (!XdmcpFill (xdmcpfd, &buf, (XdmcpNetaddr)&clnt_sa, &sa_len)) {
 	gdm_error (_("gdm_xdmcp_decode: Could not create XDMCP buffer!"));
 	return TRUE;
     }
@@ -1428,11 +1428,19 @@ gdm_xdmcp_display_alloc (const char *hostname, gint displaynum)
     d->sesspid = 0;
     d->slavepid = 0;
     d->type = TYPE_XDMCP;
+    d->console = FALSE;
     d->dispstat = XDMCP_PENDING;
     d->sessionid = globsessid++;
     d->acctime = time (NULL);
     d->dispnum = displaynum;
 
+#ifdef __linux__
+    d->vt = -1;
+#endif
+
+    d->logged_in = FALSE;
+    d->login = NULL;
+
     d->sleep_before_run = 0;
     if (GdmAllowRemoteAutoLogin &&
 	! ve_string_empty (GdmTimedLogin)) {
