@@ -78,6 +78,7 @@ gchar *desc3 = N_("This panel displays GDM's fundamental system settings.\n"
 /* Keep track of X servers, the selected user level and session details */
 int number_of_servers = 0;
 int selected_server_row = -1;
+int selected_server_def_row = -1;
 int selected_session_row = -1, default_session_row = -1;
 GtkWidget *invisible_notebook = NULL;
 GdmConfigSession *current_default_session = NULL;
@@ -85,13 +86,35 @@ GdmConfigSession *old_current_default_session = NULL;
 char *default_session_link_name = NULL;
 GList *deleted_sessions_list = NULL;
 
+const char *names_to_use[] = {
+	"foo",
+	"bar",
+	"bah",
+	"blurb",
+	"babble",
+	"bubbly",
+	"buck",
+	"bucket",
+	"cabbage",
+	"cackle",
+	"dangle",
+	"dashboard",
+	"dawn",
+	"each",
+	"earmarks",
+	"eastbound",
+	"eclipse",
+	"fabric",
+	NULL };
+
+
 /* Main application widget pointer */
 GtkWidget *GDMconfigurator = NULL;
 
 /** This is something of a hack, but it keeps the code clean and
  * easily extensible later on. 
  */
-GtkWidget *get_widget(gchar *widget_name)
+GtkWidget *get_widget(const gchar *widget_name)
 {
 	GtkWidget *ret = NULL;
 	ret = glade_xml_get_widget(GUI, widget_name);
@@ -374,6 +397,8 @@ main (int argc, char *argv[])
 				glade_filename = g_strdup ("gdmconfig.glade");
 			}
 		}
+
+		glade_helper_add_glade_directory (GDM_GLADE_DIR);
 	} else {
 		glade_filename = g_strdup ("gdmconfig.glade");
 	}
@@ -518,10 +543,155 @@ void user_level_row_selected(GtkCList *clist, gint row,
    }
 }
 
-void show_about_box(void) {
-	glade_xml_new(glade_filename, "about_gdmconfig");
+void
+show_about_box (void)
+{
+	GladeXML *xml;
+	static GtkWidget *dialog = NULL;
+
+	if (dialog != NULL) {
+		gtk_widget_show_now (dialog);
+		gdk_window_raise (dialog->window);
+		return;
+	}
+
+	xml = glade_helper_load ("gdmconfig.glade",
+				 "about_gdmconfig",
+				 gnome_about_get_type (),
+				 TRUE /* dump on destroy */);
+	dialog = glade_helper_get (xml, "about_gdmconfig",
+				   gnome_about_get_type ());
+	gnome_dialog_set_parent (GNOME_DIALOG (dialog), 
+				 (GtkWindow *) GDMconfigurator);
+	gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
+			    GTK_SIGNAL_FUNC (gtk_widget_destroyed),
+			    &dialog);
 }
 
+static int
+find_server_def (const char *id)
+{
+	GtkCList *clist = (GtkCList *) get_widget ("server_def_clist");
+	int i;
+
+	for (i = 0; i < clist->rows; i++) {
+		char *rowid = gtk_clist_get_row_data (clist, i);
+
+		if ((rowid == NULL &&
+		     id == NULL) ||
+		    (rowid != NULL &&
+		     id != NULL &&
+		     strcmp (rowid, id) == 0)) {
+			return i;
+		}
+	}
+	return -1;
+}
+
+static int
+find_default_server_def (void)
+{
+	GtkCList *clist = (GtkCList *) get_widget ("server_def_clist");
+	int i;
+
+	for (i = 0; i < clist->rows; i++) {
+		char *rowid = gtk_clist_get_row_data (clist, i);
+
+		if (rowid != NULL &&
+		    strcmp (rowid, GDM_STANDARD) == 0) {
+			return i;
+		}
+	}
+	return -1;
+}
+
+static char *
+get_unique_name (const char *except)
+{
+	int i, ii;
+
+	for (ii = 0;; ii++) {
+		for (i = 0; names_to_use [i] != NULL; i++) {
+			char *thing;
+
+			if (ii == 0)
+				thing = g_strdup (names_to_use[i]);
+			else
+				thing = g_strdup_printf ("%s%d",
+							 names_to_use[i], ii);
+			if ((except == NULL ||
+			     strcmp (except, thing) != 0) &&
+			    find_server_def (thing) < 0)
+				return thing;
+
+			g_free (thing);
+		}
+	}
+	g_assert_not_reached ();
+	/* never reached */
+	return NULL;
+}
+
+static char *
+get_server_name (const char *id)
+{
+	GtkCList *clist = (GtkCList *) get_widget ("server_def_clist");
+	int row;
+
+	row = find_server_def (id);
+	if (row >= 0) {
+		char *text = NULL;
+		gtk_clist_get_text (clist, row, 0, &text);
+
+		return g_strdup (ve_sure_string (text));
+	} else {
+		return g_strdup ("");
+	}
+}
+
+static void
+rename_row (int row, const char *to)
+{
+	GtkCList *clist = (GtkCList *) get_widget ("server_def_clist");
+	GtkCList *svr_clist = (GtkCList *) get_widget ("server_clist");
+	char *oldid;
+	char *new_name;
+	int i;
+
+	oldid = g_strdup (gtk_clist_get_row_data (clist, row));
+
+	if (to == NULL)
+		new_name = get_unique_name (oldid /* except */);
+	else
+		new_name = g_strdup (to);
+
+	if (oldid != NULL &&
+	    strcmp (oldid, new_name) == 0) {
+		g_free (new_name);
+		g_free (oldid);
+		return;
+	}
+
+	gtk_clist_set_row_data_full (clist, row, new_name,
+				     (GDestroyNotify) g_free);
+
+	for (i = 0; i < svr_clist->rows; i++) {
+		char *rowid = gtk_clist_get_row_data (clist, i);
+
+		if ((rowid == NULL &&
+		     oldid == NULL) ||
+		    (rowid != NULL &&
+		     oldid != NULL &&
+		     strcmp (rowid, oldid) == 0)) {
+			gtk_clist_set_row_data_full (svr_clist, i, new_name,
+						     (GDestroyNotify) g_free);
+		}
+	}
+
+	g_free (oldid);
+}
+
+
 void
 gdm_config_parse_most (gboolean factory)
 {
@@ -533,6 +703,9 @@ gdm_config_parse_most (gboolean factory)
     gchar *default_session_name = NULL;
     gint linklen;
     const char *config_file;
+    gboolean got_standard_server = FALSE;
+    gboolean got_any_servers = FALSE;
+    char *standard_x_server = NULL;
 
     gtk_clist_clear (GTK_CLIST (get_widget ("sessions_clist")));
     gtk_clist_clear (GTK_CLIST (get_widget ("server_clist")));
@@ -548,7 +721,8 @@ gdm_config_parse_most (gboolean factory)
      */
     if (!g_file_exists(config_file))
       {
-	      char *a_server[2];
+	      char *a_server[3];
+	      int row;
 	      char *error = g_strdup_printf (_("The configuration file: %s\n"
 					       "does not exist! Using "
 					       "default values."),
@@ -560,9 +734,15 @@ gdm_config_parse_most (gboolean factory)
 	      gnome_dialog_run_and_close(GNOME_DIALOG(error_dialog));
 
 	      a_server[0] = "0";
-	      a_server[1] = "/usr/bin/X11/X";
-	      gtk_clist_append (GTK_CLIST (get_widget ("server_clist")),
-				a_server);
+	      a_server[1] = _("Standard server");
+	      a_server[2] = "";
+	      row = gtk_clist_append (GTK_CLIST (get_widget ("server_clist")),
+				      a_server);
+	      gtk_clist_set_row_data_full
+		      (GTK_CLIST (get_widget ("server_clist")), 
+		       row,
+		       g_strdup (GDM_STANDARD),
+		       (GDestroyNotify) g_free);
 	      number_of_servers = 1;
       }
 
@@ -589,6 +769,10 @@ gdm_config_parse_most (gboolean factory)
     gdm_entry_set("session_dir", gnome_config_get_string (GDM_KEY_SESSDIR));
     gdm_entry_set("pre_session_dir", gnome_config_get_string (GDM_KEY_PRESESS));
     gdm_entry_set("post_session_dir", gnome_config_get_string (GDM_KEY_POSTSESS));
+    standard_x_server = gnome_config_get_string (GDM_KEY_STANDARD_XSERVER);
+    gdm_entry_set("standard_x_server", standard_x_server);
+    gdm_entry_set("xnest_server", gnome_config_get_string (GDM_KEY_XNEST));
+    gdm_spin_set("flexible_servers", gnome_config_get_int(GDM_KEY_FLEXIBLE_XSERVERS));
     gdm_entry_set("failsafe_x_server", gnome_config_get_string (GDM_KEY_FAILSAFE_XSERVER));
     gdm_entry_set("x_keeps_crashing", gnome_config_get_string (GDM_KEY_XKEEPSCRASHING));
     gdm_entry_set("x_keeps_crashing_configurators", gnome_config_get_string (GDM_KEY_XKEEPSCRASHING_CONFIGURATORS));
@@ -802,8 +986,6 @@ gdm_config_parse_most (gboolean factory)
 	    
 	    /*printf ("coloring session %s.\n", data->name);*/
 	    gdk_color_parse ("#d6e8ff", &col);
-	    gdk_color_alloc (gdk_colormap_get_system(), &col);
-	    
 	    gtk_clist_set_background (GTK_CLIST (get_widget("sessions_clist")),
 				      i, &col);
 	    data->is_default = TRUE;
@@ -823,29 +1005,159 @@ gdm_config_parse_most (gboolean factory)
    }
 
     gnome_config_pop_prefix();
+
+    gtk_clist_set_column_auto_resize
+	    (GTK_CLIST (get_widget ("server_def_clist")), 0, TRUE);
+    gtk_clist_set_column_auto_resize
+	    (GTK_CLIST (get_widget ("server_def_clist")), 1, TRUE);
+    gtk_clist_set_column_auto_resize
+	    (GTK_CLIST (get_widget ("server_clist")), 0, TRUE);
+    gtk_clist_set_column_auto_resize
+	    (GTK_CLIST (get_widget ("server_clist")), 1, TRUE);
+
+    got_standard_server = FALSE;
+    got_any_servers = FALSE;
+
+    /* Find server definitions */
+    iter = gnome_config_init_iterator_sections ("=" GDM_CONFIG_FILE "=/");
+    iter = gnome_config_iterator_next (iter, &key, NULL);
+    
+    while (iter) {
+	    if (strncmp (key, "server-", strlen ("server-")) == 0) {
+		    char *section;
+		    char *id;
+		    char *text[3];
+		    int row;
+
+		    section = g_strdup_printf ("=" GDM_CONFIG_FILE "=/%s/",
+					       key);
+		    gnome_config_push_prefix (section);
+
+		    id = g_strdup (key + strlen ("server-"));
+
+		    text[0] = gnome_config_get_string (GDM_KEY_SERVER_NAME);
+		    if (text[0] == NULL)
+			    text[0] = g_strdup (id);
+		    text[1] = gnome_config_get_string (GDM_KEY_SERVER_COMMAND);
+		    if (text[1] == NULL)
+			    text[1] = g_strdup (standard_x_server);
+		    if (text[1] == NULL)
+			    text[1] = g_strdup ("/usr/bin/X11/X");
+
+		    if (gnome_config_get_bool (GDM_KEY_SERVER_FLEXIBLE))
+			    text[2] = g_strdup (_("Yes"));
+		    else
+			    text[2] = g_strdup (_("No"));
+
+		    g_free (section);
+		    gnome_config_pop_prefix ();
+
+		    row = gtk_clist_append
+			    (GTK_CLIST (get_widget ("server_def_clist")), text);
+		    gtk_clist_set_row_data_full
+			    (GTK_CLIST (get_widget ("server_def_clist")), 
+			     row,
+			     id,
+			     (GDestroyNotify) g_free);
+
+		    g_free (text[0]);
+		    g_free (text[1]);
+		    g_free (text[2]);
+
+		    got_any_servers = TRUE;
+		    if (strcmp (id, GDM_STANDARD) == 0) {
+			    GdkColor col;
+			    gdk_color_parse ("#d6e8ff", &col);
+			    gtk_clist_set_background
+				    (GTK_CLIST (get_widget("server_def_clist")),
+				     row, &col);
+			    got_standard_server = TRUE;
+		    }
+	    }
+
+	    g_free (key);
+
+	    iter = gnome_config_iterator_next (iter, &key, NULL);
+    }
+
+    if ( ! got_any_servers ||
+	 ! got_standard_server) {
+	    char *text[3];
+	    int row;
+	    GdkColor col;
+
+	    text[0] = g_strdup (_("Standard server"));
+	    text[1] = g_strdup (standard_x_server);
+	    if (text[1] == NULL)
+		    text[1] = g_strdup ("/usr/bin/X11/X");
+
+	    text[2] = g_strdup (_("Yes"));
+
+	    row = gtk_clist_append
+		    (GTK_CLIST (get_widget ("server_def_clist")), text);
+	    gtk_clist_set_row_data_full
+		    (GTK_CLIST (get_widget ("server_def_clist")), 
+		     row,
+		     g_strdup (GDM_STANDARD),
+		     (GDestroyNotify) g_free);
+
+	    gdk_color_parse ("#d6e8ff", &col);
+	    gtk_clist_set_background
+		    (GTK_CLIST (get_widget("server_def_clist")), row, &col);
+
+	    g_free (text[0]);
+	    g_free (text[1]);
+	    g_free (text[2]);
+    }
     
     /* Fill the widgets in Servers tab */
     iter=gnome_config_init_iterator("=" GDM_CONFIG_FILE "=/" GDM_KEY_SERVERS);
     iter=gnome_config_iterator_next (iter, &key, &value);
     
     while (iter) {
-	char *a_server[2];
+	char *a_server[3];
 	
-	if(isdigit(*key)) 
-	  {
-	      a_server[0] = key;
-	      a_server[1] = value;
-	      gtk_clist_append(GTK_CLIST(get_widget("server_clist")), a_server);
-	      number_of_servers++;
-	  }
-	else
-	  {
+	if (isdigit (*key)) {
+		char *first = ve_first_word (ve_sure_string (value));
+		a_server[0] = key;
+		if (first[0] == '/') {
+			a_server[1] = value;
+			a_server[2] = "";
+			gtk_clist_append
+				(GTK_CLIST (get_widget ("server_clist")),
+				 a_server);
+			g_free (first);
+		} else {
+			int row;
+			char *rest = ve_rest (value);
+			char *name = get_server_name (first);
+			if (rest == NULL)
+				rest = g_strdup ("");
+			a_server[1] = name;
+			a_server[2] = rest;
+			row = gtk_clist_append
+				(GTK_CLIST (get_widget ("server_clist")),
+				 a_server);
+			gtk_clist_set_row_data_full
+				(GTK_CLIST (get_widget ("server_clist")),
+				 row,
+				 first,
+				 (GDestroyNotify) g_free);
+			g_free (name);
+			g_free (rest);
+		}
+		number_of_servers++;
+	} else {
 	      gnome_warning_dialog_parented(_("gdm_config_parse_most: Invalid server line in config file. Ignoring!"),
 					    GTK_WINDOW(GDMconfigurator));
-	  }
+	}
+	g_free (key);
+	g_free (value);
 	iter=gnome_config_iterator_next (iter, &key, &value);
     }
     /* FIXME: Could do something nicer with the 'exclude users' GUI sometime */
+
+    g_free (standard_x_server);
 }
 
 
@@ -990,6 +1302,7 @@ run_warn_reset_dialog (void)
 		       "take effect until gdm is restarted or the computer is\n"
 		       "rebooted"),
 		     GTK_WINDOW (GDMconfigurator));
+		gtk_window_set_modal (GTK_WINDOW (w), TRUE);
 		gnome_dialog_run_and_close (GNOME_DIALOG (w));
 	}
 	/* Don't apply identical settings again. */
@@ -1360,6 +1673,7 @@ write_config (void)
 				"saved.\n"));
 	    dlg = gnome_error_dialog_parented (errors->str,
 					       GTK_WINDOW(GDMconfigurator));
+	    gtk_window_set_modal (GTK_WINDOW (dlg), TRUE);
 	    gnome_dialog_run_and_close (GNOME_DIALOG (dlg));
 	    g_string_free (errors, TRUE);
     }
@@ -1524,62 +1838,542 @@ change_timed_sensitivity (GtkButton *button,
 				  GTK_TOGGLE_BUTTON (button)->active);
 }
 
-void handle_server_add_or_edit         (gchar           *string,
-					gpointer         user_data)
+static void
+handle_server_def_edit (gboolean edit)
 {
-    if(!string)
-      return;
+	GladeXML *xml;
+	GtkWidget *dialog, *name, *command_line, *flexible, *standard;
+	GtkCList *server_def_clist;
+	char *current_name, *current_command, *current_flexible, *current_id;
 
-    /* Add a new server to the end of the server CList */
-    if (strcmp("add", (char *)user_data) == 0)
-      {
-	  gchar *new_server[2];
+	if (edit && selected_server_def_row < 0)
+		return;
 
-	  new_server[0] = g_strdup_printf("%d", number_of_servers);;
-	  new_server[1] = string;
-	  /* We now have an extra server */
-	  number_of_servers++;
-	  gtk_clist_append(GTK_CLIST(get_widget("server_clist")), new_server);
-      }
-    else
-      {
-	  /* Update the current server */
-	  gtk_clist_set_text(GTK_CLIST(get_widget("server_clist")),
-			     selected_server_row, 1,
-			     string);
-      }
+	server_def_clist = GTK_CLIST (get_widget ("server_def_clist"));
+
+	xml = glade_helper_load ("gdmconfig.glade", "svr_def",
+				 gnome_dialog_get_type (),
+				 TRUE /* dump on destroy */);
+	dialog = glade_helper_get (xml, "svr_def",
+				   gnome_dialog_get_type ());
+	name = glade_helper_get (xml, "svr_def_name",
+				 gtk_entry_get_type ());
+	command_line = glade_helper_get (xml, "svr_def_command_line",
+					 gtk_entry_get_type ());
+	flexible = glade_helper_get (xml, "svr_def_flexible",
+				     gtk_toggle_button_get_type ());
+	standard = glade_helper_get (xml, "svr_def_standard",
+				     gtk_toggle_button_get_type ());
+
+	if (edit) {
+		gtk_clist_get_text (server_def_clist,
+				    selected_server_def_row, 0,
+				    &current_name);
+		gtk_clist_get_text (server_def_clist,
+				    selected_server_def_row, 1,
+				    &current_command);
+		gtk_clist_get_text (server_def_clist,
+				    selected_server_def_row, 2,
+				    &current_flexible);
+		current_id = gtk_clist_get_row_data (server_def_clist,
+						     selected_server_def_row);
+
+		gtk_entry_set_text (GTK_ENTRY (name),
+				    ve_sure_string (current_name));
+		gtk_entry_set_text (GTK_ENTRY (command_line),
+				    ve_sure_string (current_command));
+		if (strcmp (_("Yes"), current_flexible) == 0)
+			gtk_toggle_button_set_active
+				(GTK_TOGGLE_BUTTON (flexible), TRUE);
+		else
+			gtk_toggle_button_set_active
+				(GTK_TOGGLE_BUTTON (flexible), FALSE);
+		if (current_id == NULL ||
+		    strcmp (current_id, GDM_STANDARD) == 0) {
+			gtk_toggle_button_set_active
+				(GTK_TOGGLE_BUTTON (standard), TRUE);
+		} else {
+			gtk_toggle_button_set_active
+				(GTK_TOGGLE_BUTTON (standard), FALSE);
+		}
+	} else {
+		gtk_entry_set_text (GTK_ENTRY (name),
+				    _("Standard server"));
+		gtk_entry_set_text (GTK_ENTRY (command_line),
+				    "/usr/bin/X11/X");
+		gtk_toggle_button_set_active
+			(GTK_TOGGLE_BUTTON (flexible), TRUE);
+		gtk_toggle_button_set_active
+			(GTK_TOGGLE_BUTTON (standard), TRUE);
+	}
+
+	gnome_dialog_set_parent (GNOME_DIALOG (dialog), 
+				 (GtkWindow *) GDMconfigurator);
+	gnome_dialog_close_hides (GNOME_DIALOG (dialog), TRUE);
+	gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
+
+	for (;;) {
+		if (gnome_dialog_run (GNOME_DIALOG (dialog)) == 0) {
+			char *errmsg = NULL;
+			char *cmdline = gtk_entry_get_text
+					(GTK_ENTRY (command_line));
+			char *thename = gtk_entry_get_text (GTK_ENTRY (name));
+			if (cmdline[0] != '/') {
+				errmsg = _("A command line must start "
+					   "with a forward slash "
+					   "('/')");
+			} else if (ve_string_empty (thename)) {
+				errmsg = _("A descriptive server name must "
+					   "be supplied");
+			}
+
+			if (errmsg != NULL) {
+				GtkWidget *error_dialog =
+					gnome_error_dialog (errmsg);
+				gnome_dialog_set_parent
+					(GNOME_DIALOG (error_dialog), 
+					 (GtkWindow *) dialog);
+				gtk_window_set_modal
+					(GTK_WINDOW (error_dialog),
+					 TRUE);
+				gnome_dialog_run_and_close
+					(GNOME_DIALOG (error_dialog));
+				continue;
+			}
+			break;
+		} else {
+			gtk_widget_destroy (dialog);
+			return;
+		}
+	}
+
+	current_name = gtk_entry_get_text (GTK_ENTRY (name));
+	current_command = gtk_entry_get_text (GTK_ENTRY (command_line));
+
+	if (edit) {
+		gtk_clist_set_text (server_def_clist,
+				    selected_server_def_row, 0,
+				    ve_sure_string (current_name));
+		gtk_clist_set_text (server_def_clist,
+				    selected_server_def_row, 1,
+				    ve_sure_string (current_command));
+		if (GTK_TOGGLE_BUTTON (flexible)->active)
+			gtk_clist_set_text
+				(GTK_CLIST (get_widget ("server_def_clist")),
+				 selected_server_def_row, 2, _("Yes"));
+		else
+			gtk_clist_set_text
+				(GTK_CLIST (get_widget ("server_def_clist")),
+				 selected_server_def_row, 2, _("No"));
+
+		if (GTK_TOGGLE_BUTTON (standard)->active) {
+			make_server_def_default (NULL, NULL);
+		}
+	} else {
+		int row;
+		char *server[3];
+
+		server[0] = current_name;
+		server[1] = current_command;
+		if (GTK_TOGGLE_BUTTON (flexible)->active)
+			server[2] = _("Yes");
+		else
+			server[2] = _("No");
+
+		row = gtk_clist_append (server_def_clist, server);
+
+		if (GTK_TOGGLE_BUTTON (standard)->active) {
+			/* UGLY! */
+			int tmp = selected_server_def_row;
+			selected_server_def_row = row;
+			make_server_def_default (NULL, NULL);
+			selected_server_def_row = tmp;
+		} else {
+			char *id = get_unique_name (NULL);
+			gtk_clist_set_row_data_full (server_def_clist, row, id,
+						     (GDestroyNotify) g_free);
+		}
+	}
+
+	gtk_widget_destroy (dialog);
+}
+
+void
+add_new_server_def (GtkButton *button, gpointer user_data)
+{
+	handle_server_def_edit (FALSE /*edit*/);
 }
 
 
 void
-add_new_server                         (GtkButton       *button,
-                                        gpointer         user_data)
+edit_selected_server_def (GtkButton *button, gpointer user_data)
 {
-    /* Request the command line for this new server */
-    gnome_request_dialog(FALSE, _("Enter the path to the X server,and\nany parameters that should be passed to it."),
-			 "/usr/bin/X11/X",
-			 -1, handle_server_add_or_edit,
-			 (gpointer)"add", (GtkWindow *)GDMconfigurator);
+	handle_server_def_edit (TRUE /*edit*/);
 }
 
+void
+delete_selected_server_def (GtkButton *button, gpointer user_data)
+{
+	int tmp = selected_server_def_row;
+	int i;
+	char *id;
+	char *cmdline = NULL;
+	GtkCList *server_def_clist;
+	GtkCList *server_clist;
+
+	if (selected_server_def_row < 0)
+		return;
+
+	server_clist = GTK_CLIST (get_widget ("server_clist"));
+	server_def_clist = GTK_CLIST (get_widget ("server_def_clist"));
+
+	id = gtk_clist_get_row_data (server_def_clist, tmp);
+	gtk_clist_get_text (server_def_clist, tmp, 1, &cmdline);
+
+	for (i = 0; i < server_clist->rows; i++) {
+		char *rowid = gtk_clist_get_row_data (server_clist, i);
+		if ((rowid == NULL && id == NULL) ||
+		    (rowid != NULL && id != NULL && strcmp (rowid, id) == 0)) {
+			char *newcmd;
+			char *extra = NULL;
+
+			gtk_clist_get_text (server_clist, i, 2, &extra);
+			if (ve_string_empty (extra))
+				newcmd = g_strdup (ve_sure_string (cmdline));
+			else
+				newcmd = g_strconcat (ve_sure_string (cmdline),
+						      " ", extra, NULL);
+
+			gtk_clist_set_text (server_clist, i, 1, newcmd);
+			gtk_clist_set_text (server_clist, i, 2, "");
+
+			g_free (newcmd);
+		}
+	}
+
+	/* Remove the server from the list */
+	selected_server_def_row = -1;
+	gtk_clist_remove (server_def_clist, tmp);
+}
 
 void
-edit_selected_server                   (GtkButton       *button,
-                                        gpointer         user_data)
+make_server_def_default (GtkButton *button, gpointer user_data)
 {
-    char *current_server;
+	int old_dfl;
+	GdkColor col;
+
+	if (selected_server_def_row < 0)
+		return;
+
+	old_dfl = find_default_server_def ();
+
+	if (old_dfl == selected_server_def_row)
+		return;
+
+	if (old_dfl >= 0) {
+		rename_row (old_dfl, NULL);
+		gtk_clist_set_background
+			(GTK_CLIST (get_widget ("server_def_clist")),
+			 old_dfl, NULL);
+	}
+	rename_row (selected_server_def_row, GDM_STANDARD);
+
+	gdk_color_parse ("#d6e8ff", &col);
+	gtk_clist_set_background (GTK_CLIST (get_widget ("server_def_clist")),
+				  selected_server_def_row, &col);
+}
+
+void
+record_selected_server_def (GtkCList *clist,
+			    gint row,
+			    gint column,
+			    GdkEventButton *event,
+			    gpointer user_data)
+{
+	selected_server_def_row = row;
+}
+
+static void
+fill_svr_select_list (GladeXML *xml)
+{
+	int i;
+	GtkCList *svr_clist = (GtkCList *)get_widget ("server_def_clist");
+	GtkCList *clist =
+		(GtkCList *)glade_helper_get_clist (xml,
+						    "svr_select_clist",
+						    gtk_clist_get_type (),
+						    1);
+	for (i = 0; i < svr_clist->rows; i++) {
+		int row;
+		char *text = NULL;
+		char *id = NULL;
+		gtk_clist_get_text (svr_clist, i, 0, &text);
+		id = gtk_clist_get_row_data (svr_clist, i);
+		if (text != NULL) {
+			row = gtk_clist_append (clist, &text);
+			gtk_clist_set_row_data_full
+				(clist, row,
+				 g_strdup (id),
+				 (GDestroyNotify) g_free);
+		}
+	}
+
+	if (i == 0) {
+		int row;
+		char *text = _("Standard server");
+		row = gtk_clist_append (clist, &text);
+		gtk_clist_set_row_data_full
+			(clist, row,
+			 g_strdup (GDM_STANDARD),
+			 (GDestroyNotify) g_free);
+	}
+}
+
+static void
+custom_toggled (GtkWidget *w, gpointer data)
+{
+	GladeXML *xml = data;
+	GtkWidget *clist = glade_helper_get_clist (xml, "svr_select_clist",
+						   gtk_clist_get_type (), 1);
+	GtkWidget *cmdline = glade_helper_get (xml, "svr_select_command_line",
+					       gtk_entry_get_type ());
+	GtkWidget *extra = glade_helper_get (xml, "svr_select_arguments",
+					     gtk_entry_get_type ());
+
+	if (GTK_TOGGLE_BUTTON (w)->active) {
+		gtk_widget_set_sensitive (clist, FALSE);
+		gtk_widget_set_sensitive (extra, FALSE);
+		gtk_widget_set_sensitive (cmdline, TRUE);
+	} else {
+		gtk_widget_set_sensitive (clist, TRUE);
+		gtk_widget_set_sensitive (extra, TRUE);
+		gtk_widget_set_sensitive (cmdline, FALSE);
+	}
+}
+
+static void
+connect_custom (GladeXML *xml)
+{
+	GtkWidget *button = glade_helper_get (xml, "svr_select_custom_cb",
+					      gtk_toggle_button_get_type ());
+
+	gtk_object_ref (GTK_OBJECT (xml));
+	gtk_signal_connect_full (GTK_OBJECT (button), "toggled",
+				 GTK_SIGNAL_FUNC (custom_toggled),
+				 NULL /* marshal*/,
+				 xml,
+				 (GtkDestroyNotify) gtk_object_unref,
+				 FALSE /*object_signal*/,
+				 FALSE /*after*/);
+
+	/* setup sensitivities */
+	custom_toggled (button, xml);
+}
+
+static void
+select_svr (GtkCList *clist, const char *current_id)
+{
+	int i;
+
+	for (i = 0; i < clist->rows; i++) {
+		char *id = NULL;
+		id = gtk_clist_get_row_data (clist, i);
+		if (id != NULL && current_id != NULL &&
+		    strcmp (id, current_id) == 0) {
+			gtk_clist_select_row (clist, i, 0);
+			return;
+		}
+	}
+}
+
+static void
+handle_server_edit (gboolean edit)
+{
+	GladeXML *xml;
+	GtkWidget *dialog, *custom_cb, *clist, *cmdline, *extra;
+	char *current_server = NULL, *current_extra = NULL;
+	char *current_id;
     
-    if (selected_server_row < 0)
-      return;
+	if (edit && selected_server_row < 0)
+		return;
 
-    /* Ask for the new name of this server */
-    gtk_clist_get_text(GTK_CLIST(get_widget("server_clist")),
-		       selected_server_row, 1, &current_server);
-    gnome_request_dialog(FALSE, _("Enter the path to the X server,and\nany parameters that should be passed to it."),
-			 current_server,
-			 -1, handle_server_add_or_edit,
-			 (gpointer)"edit", (GtkWindow *)GDMconfigurator);
+	xml = glade_helper_load ("gdmconfig.glade", "svr_select",
+				 gnome_dialog_get_type (),
+				 TRUE /* dump on destroy */);
+	dialog = glade_helper_get (xml, "svr_select",
+				   gnome_dialog_get_type ());
+	custom_cb = glade_helper_get (xml, "svr_select_custom_cb",
+				      gtk_toggle_button_get_type ());
+	clist = glade_helper_get_clist (xml, "svr_select_clist",
+					gtk_clist_get_type (), 1);
+	cmdline = glade_helper_get (xml, "svr_select_command_line",
+				    gtk_entry_get_type ());
+	extra = glade_helper_get (xml, "svr_select_arguments",
+				  gtk_entry_get_type ());
 
+	fill_svr_select_list (xml);
+	connect_custom (xml);
+
+	gtk_entry_set_text (GTK_ENTRY (cmdline), "/usr/bin/X11/X");
+
+	if (edit) {
+		gtk_clist_get_text (GTK_CLIST (get_widget ("server_clist")),
+				    selected_server_row, 1, &current_server);
+		gtk_clist_get_text (GTK_CLIST (get_widget ("server_clist")),
+				    selected_server_row, 2, &current_extra);
+		current_id = gtk_clist_get_row_data
+			(GTK_CLIST (get_widget ("server_clist")),
+			 selected_server_row);
+
+		gtk_entry_set_text (GTK_ENTRY (extra),
+				    ve_sure_string (current_extra));
+		if (current_id == NULL) {
+			gtk_toggle_button_set_active
+				(GTK_TOGGLE_BUTTON (custom_cb), TRUE);
+			gtk_entry_set_text (GTK_ENTRY (cmdline),
+					    ve_sure_string (current_server));
+		} else {
+			select_svr (GTK_CLIST (clist), current_id);
+		}
+	}
+
+	gnome_dialog_set_parent (GNOME_DIALOG (dialog), 
+				 (GtkWindow *) GDMconfigurator);
+	gnome_dialog_close_hides (GNOME_DIALOG (dialog), TRUE);
+	gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
+
+	for (;;) {
+		if (gnome_dialog_run (GNOME_DIALOG (dialog)) == 0) {
+			if (GTK_TOGGLE_BUTTON (custom_cb)->active) {
+				char *server = gtk_entry_get_text
+					(GTK_ENTRY (cmdline));
+				if (server[0] != '/') {
+					GtkWidget *error_dialog =
+						gnome_error_dialog
+						(_("A command line must start "
+						   "with a forward slash "
+						   "('/')"));
+					gnome_dialog_set_parent
+						(GNOME_DIALOG (error_dialog), 
+						 (GtkWindow *) dialog);
+					gtk_window_set_modal
+						(GTK_WINDOW (error_dialog),
+						 TRUE);
+					gnome_dialog_run_and_close
+						(GNOME_DIALOG (error_dialog));
+					continue;
+				}
+			} else {
+				if (GTK_CLIST (clist)->selection == NULL)
+					continue;
+			}
+			break;
+		} else {
+			gtk_widget_destroy (dialog);
+			return;
+		}
+	}
+
+	if (edit) {
+		if (GTK_TOGGLE_BUTTON (custom_cb)->active) {
+			char *server = gtk_entry_get_text (GTK_ENTRY (cmdline));
+			gtk_clist_set_text
+				(GTK_CLIST (get_widget ("server_clist")),
+				 selected_server_row, 1, server);
+			gtk_clist_set_text
+				(GTK_CLIST (get_widget ("server_clist")),
+				 selected_server_row, 2, "");
+			gtk_clist_set_row_data
+				(GTK_CLIST (get_widget ("server_clist")),
+				 selected_server_row, NULL);
+		} else {
+			char *args = gtk_entry_get_text (GTK_ENTRY (extra));
+			char *server, *id;
+			int sel = GPOINTER_TO_INT
+				(GTK_CLIST (clist)->selection->data);
+			server = NULL;
+			gtk_clist_get_text (GTK_CLIST (clist), sel, 0, &server);
+			id = gtk_clist_get_row_data (GTK_CLIST (clist), sel);
+
+			gtk_clist_set_text
+				(GTK_CLIST (get_widget ("server_clist")),
+				 selected_server_row, 1,
+				 ve_sure_string (server));
+			gtk_clist_set_text
+				(GTK_CLIST (get_widget ("server_clist")),
+				 selected_server_row, 2,
+				 ve_sure_string (args));
+			gtk_clist_set_row_data_full
+				(GTK_CLIST (get_widget ("server_clist")),
+				 selected_server_row, 
+				 g_strdup (id),
+				 (GDestroyNotify) g_free);
+		}
+	} else {
+		if (GTK_TOGGLE_BUTTON (custom_cb)->active) {
+			char *server = gtk_entry_get_text (GTK_ENTRY (cmdline));
+			char *new_server[3];
+
+			new_server[0] =
+				g_strdup_printf("%d", number_of_servers);
+			new_server[1] = server;
+			new_server[2] = "";
+
+			/* We now have an extra server */
+			number_of_servers++;
+
+			gtk_clist_append(GTK_CLIST(get_widget("server_clist")),
+					 new_server);
+
+			g_free (new_server[0]);
+		} else {
+			char *new_server[3];
+			char *args = gtk_entry_get_text (GTK_ENTRY (extra));
+			char *server, *id;
+			int row;
+			int sel = GPOINTER_TO_INT
+				(GTK_CLIST (clist)->selection->data);
+			server = NULL;
+			gtk_clist_get_text (GTK_CLIST (clist), sel, 0, &server);
+			id = gtk_clist_get_row_data (GTK_CLIST (clist), sel);
+
+
+			new_server[0] =
+				g_strdup_printf("%d", number_of_servers);
+			new_server[1] = ve_sure_string (server);
+			new_server[2] = ve_sure_string (args);
+
+			/* We now have an extra server */
+			number_of_servers++;
+
+			row = gtk_clist_append
+				(GTK_CLIST (get_widget ("server_clist")),
+				 new_server);
+
+			g_free (new_server[0]);
+
+
+			gtk_clist_set_row_data_full
+				(GTK_CLIST (get_widget ("server_clist")),
+				 row, 
+				 g_strdup (id),
+				 (GDestroyNotify) g_free);
+		}
+	}
+
+	gtk_widget_destroy (dialog);
+}
+
+void
+add_new_server (GtkButton *button, gpointer user_data)
+{
+	handle_server_edit (FALSE /*edit*/);
+}
+
+void
+edit_selected_server (GtkButton *button, gpointer user_data)
+{
+	handle_server_edit (TRUE /*edit*/);
 }
 
 
@@ -1592,30 +2386,32 @@ delete_selected_server                 (GtkButton       *button,
     if (selected_server_row < 0)
       return;
 
+    selected_server_row = -1;
+
     /* Remove the server from the list */
     gtk_clist_remove(GTK_CLIST(get_widget("server_clist")), 
-		     selected_server_row);
+		     tmp);
     number_of_servers--;
     
     /* The server numbers will be out of sync now, so correct that */ 
-    while (tmp < number_of_servers)
-      {
-	  gtk_clist_set_text(GTK_CLIST(get_widget("server_clist")), tmp, 0,
-			     g_strdup_printf("%d", tmp));
-	  tmp++;
-      }
-    selected_server_row = -1;
+    /* FIXME: this is not right */
+    while (tmp < number_of_servers) {
+	      char *num = g_strdup_printf("%d", tmp);
+	      gtk_clist_set_text (GTK_CLIST (get_widget("server_clist")),
+				  tmp, 0, num);
+	      tmp++;
+    }
 }
 
 
 void
-record_selected_server                  (GtkCList *clist,
-			        	 gint row,
-					 gint column,
-					 GdkEventButton *event,
-					 gpointer user_data)
+record_selected_server (GtkCList *clist,
+			gint row,
+			gint column,
+			GdkEventButton *event,
+			gpointer user_data)
 {
-    selected_server_row = row;
+	selected_server_row = row;
 }
 
 
@@ -1737,7 +2533,6 @@ set_new_default_session (GtkButton *button,
    current_default_session->is_default = TRUE;
    
    gdk_color_parse ("#d6e8ff", &col);
-   gdk_color_alloc (gdk_colormap_get_system(), &col);
    gtk_clist_freeze (GTK_CLIST (get_widget("sessions_clist")));
 
    gtk_clist_set_background (GTK_CLIST (get_widget("sessions_clist")),
@@ -1796,6 +2591,7 @@ add_session_real (gchar *new_session_name, gpointer data)
 					 "not empty"));
 	   gnome_dialog_set_parent (GNOME_DIALOG (error_dialog), 
 				   (GtkWindow *) GDMconfigurator);
+	   gtk_window_set_modal (GTK_WINDOW (error_dialog), TRUE);
 	   gnome_dialog_run_and_close (GNOME_DIALOG (error_dialog));
    }
 }
