@@ -187,6 +187,79 @@ ve_first_word (const char *s)
 	return ret;
 }
 
+char *
+ve_rest (const char *s)
+{
+	const char *p;
+	gboolean single_quot = FALSE;
+	gboolean double_quot = FALSE;
+	gboolean escape = FALSE;
+
+	if (s == NULL)
+		return NULL;
+
+	for (p = s; *p != '\0'; p++) {
+		if (single_quot) {
+			if (*p == '\'') {
+				single_quot = FALSE;
+			}
+		} else if (escape) {
+			escape = FALSE;
+		} else if (double_quot) {
+			if (*p == '"') {
+				double_quot = FALSE;
+			} else if (*p == '\\') {
+				escape = TRUE;
+			}
+		} else if (*p == '\'') {
+			single_quot = TRUE;
+		} else if (*p == '"') {
+			double_quot = TRUE;
+		} else if (*p == '\\') {
+			escape = TRUE;
+		} else if (*p == ' ' || *p == '\t') {
+			while (*p == ' ' || *p == '\t')
+				p++;
+			return g_strdup (p);
+		}
+	}
+
+	return NULL;
+}
+
+char **
+ve_vector_merge (char * const *v1, char * const *v2)
+{
+	int len1, argc, i;
+	char **argv;
+
+	if (v1 == NULL && v2 == NULL)
+		return NULL;
+
+	len1 = ve_vector_len (v1);
+	argc = len1 + ve_vector_len (v2);
+
+	argv = g_new (char *, argc+1);
+	for (i = 0; i < len1; i++)
+		argv[i] = g_strdup (v1[i]);
+	for (; i < argc; i++)
+		argv[i] = g_strdup (v2[i - len1]);
+	argv[i] = NULL;
+
+	return argv;
+}
+
+int
+ve_vector_len (char * const *v)
+{
+	int i;
+	if (v == NULL)
+		return 0;
+	for (i = 0; v[i] != NULL; i++)
+		;
+	return i;
+}
+
 gboolean
 ve_is_string_in_list (const GList *list, const char *string)
 {
